Package: aftables
Title: Create Spreadsheet Publications Following Best Practice
Version: 2.0.1
Authors@R: c(
    person("Matt", "Dray", , "mwdray@gmail.com", role = c("aut")),
    person("Tim", "Taylor", role = "ctb"),
    person("Matt", "Kerlogue", role = "ctb"),
    person("Crown Copyright", role = c("cph")),
    person("Olivia", "Box Power", , "Olivia.BoxPower@dhsc.gov.uk", role = c("cre", "ctb")),
    person("Zachary", "Gleisner", , "Zachary.Gleisner@dhsc.gov.uk", role = "ctb")
  )
Description: Generate spreadsheet publications that follow best practice
    guidance from the UK government's Analysis Function, available at 
    <https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/>,
    with a focus on accessibility. See also the 'Python' package 'gptables'.
License: MIT + file LICENSE
URL: https://best-practice-and-impact.github.io/aftables/,
        https://github.com/best-practice-and-impact/aftables
BugReports: https://github.com/best-practice-and-impact/aftables/issues
Depends: R (>= 4.1.0)
Imports: openxlsx2, pillar, purrr, dplyr, stringr, tidyselect, tidyr,
        rlang, utils, yaml, tibble
Suggests: covr, knitr, rmarkdown, rstudioapi, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-22 12:00:07 UTC; OBoxPower
Author: Matt Dray [aut],
  Tim Taylor [ctb],
  Matt Kerlogue [ctb],
  Crown Copyright [cph],
  Olivia Box Power [cre, ctb],
  Zachary Gleisner [ctb]
Maintainer: Olivia Box Power <Olivia.BoxPower@dhsc.gov.uk>
Repository: CRAN
Date/Publication: 2026-04-22 12:10:02 UTC
Built: R 4.5.2; ; 2026-04-22 13:01:37 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(summary,aftable)
S3method(tbl_sum,aftable)
export(as_aftable)
export(at_template_aftable)
export(at_template_workflow)
export(create_aftable)
export(create_config_yaml)
export(generate_workbook)
export(is_aftable)
import(openxlsx2)
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,mutate)
importFrom(pillar,tbl_sum)
importFrom(purrr,compact)
importFrom(purrr,list_modify)
importFrom(purrr,map_depth)
importFrom(purrr,pluck)
importFrom(purrr,pluck_exists)
importFrom(purrr,pwalk)
importFrom(rlang,'%||%')
importFrom(rlang,.data)
importFrom(stringr,str_extract)
importFrom(stringr,str_replace)
importFrom(stringr,str_replace_all)
importFrom(tibble,tibble)
importFrom(tidyselect,where)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# {aftables} <a href="https://best-practice-and-impact.github.io/aftables/"><img src="man/figures/logo.png" align="right" height="139"/></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/aftables)](https://CRAN.R-project.org/package=aftables)
[![R-CMD-check](https://github.com/best-practice-and-impact/aftables/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/best-practice-and-impact/aftables/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

## Purpose

An R package to help automatically create reproducible spreadsheets that
adhere to the guidance on [releasing statistics in
spreadsheets](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/)
from the UK government’s [Analysis
Function](https://analysisfunction.civilservice.gov.uk/), with a focus
on accessibility.

Visit [the {aftables}
website](https://best-practice-and-impact.github.io/aftables/) for
documentation.

## Accessibility

This package is not yet capable of creating perfectly accessible
spreadsheets but will help with the bulk of the work needed. Users of
this packages should refer back to [the main spreadsheet
guidance](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/)
or [the spreadsheet accessibility
checklist](https://analysisfunction.civilservice.gov.uk/policy-store/making-spreadsheets-accessible-a-brief-checklist-of-the-basics/)
after using it to make sure nothing has been missed. Please email
<analysis.function@ons.gov.uk> if you use the package so they can
monitor use and the outputs produced.

## Contribute

The package is under (opinionated) active development. Please see [the
NEWS
file](https://best-practice-and-impact.github.io/aftables/news/index.html)
for the latest changes.

To contribute, please add [an
issue](https://github.com/best-practice-and-impact/aftables/issues) or
[a pull
request](https://github.com/best-practice-and-impact/aftables/pulls)
after reading [the code of
conduct](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CODE_OF_CONDUCT.md)
and
[contributing](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CONTRIBUTING.md)
guidance.

## Installation

### Install from CRAN

Install the latest release version of aftables directly from CRAN:

``` r
install.packages("aftables")
```

### Install from GitHub

aftables can be installed directly from GitHub.

``` r
remotes::install_github(
  "best-practice-and-impact/aftables",
  upgrade = "never",
  build_vignettes = TRUE,
  dependencies = TRUE
)
```

## How to install a11ytables

The original name of the aftables package was a11ytables. The package
was renamed to make it consistent the [afcharts
package](https://github.com/best-practice-and-impact/afcharts). Some
function names were updated when the package was renamed. If you need to
install a11ytables for a legacy project, the following code can be used.

``` r
install.packages("remotes") # if not already installed

remotes::install_github(
  repo = "best-practice-and-impact/aftables", # GitHub user/repository
  ref = "v0.3.2", # this installs the final version of a11ytables. Change if you need an earlier version.
  dependencies = TRUE, # install required/suggested packages
  build_vignettes = TRUE # generate vignette documentation
)

library(a11ytables) # attach package
```

## Use

To create a spreadsheet:

1.  Use `create_aftable()`
2.  Pass the output to `generate_workbook()`
3.  Pass the output to `openxlsx2::wb_save()`

Run `?function_name` or visit [the package
website](https://best-practice-and-impact.github.io/aftables/reference/index.html)
for function documentation. For long-form documentation, [visit the
package website](https://best-practice-and-impact.github.io/aftables/)
or run `browseVignettes("aftables")` to read the:

- [introductory
  vignette](https://best-practice-and-impact.github.io/aftables/articles/aftables.html)
  to get started
- [accessibility checklist
  vignette](https://best-practice-and-impact.github.io/aftables/articles/checklist.html)
  to see how the package complies with best-practice guidance
- [terminology
  vignette](https://best-practice-and-impact.github.io/aftables/articles/terminology.html)
  to understand the nomenclature of spreadsheet terms as used in this
  package
- [package structure
  vignette](https://best-practice-and-impact.github.io/aftables/articles/structure.html)
  to see how the package works under the hood

This package also includes [an RStudio
Addin](https://rstudio.github.io/rstudioaddins/) that inserts pre-filled
demo skeletons of the {aftables} workflow.

## Related projects

The ONS’s Analysis Standards and Pipelines team has released [a Python
package called ‘gptables’](https://github.com/ONSdigital/gptables).
{aftables} is an independent effort that offers a native R solution that
is very similar to gptables in its outputs, though there are some
differences in implementation.

{aftables} can help you fulfill a [Reproducible Analytical
Pipeline](https://analysisfunction.civilservice.gov.uk/support/reproducible-analytical-pipelines/)
by automating the generation of compliant spreadsheets for publication.

## Code of Conduct

Please note that the {aftables} project is released with a [Contributor
Code of
Conduct](https://best-practice-and-impact.github.io/aftables/CODE_OF_CONDUCT.html).

## Copyright and Licensing

© Crown Copyright, 2026.

This work is [Crown
Copyright](https://www.nationalarchives.gov.uk/information-management/re-using-public-sector-information/uk-government-licensing-framework/crown-copyright/).
The source code for the software is released under the MIT licence as
per the the [UK Government Licensing
Framework](https://www.nationalarchives.gov.uk/information-management/re-using-public-sector-information/uk-government-licensing-framework/open-government-licence/open-software-licences/)
and the [GDS Way licensing
guidance](https://gds-way.digital.cabinet-office.gov.uk/manuals/licensing.html).
The documentation for the software is released under the [Open
Government
Licence](https://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/).

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)

## ----results='asis', echo = FALSE---------------------------------------------
cat("```xml\n")
cat(readLines(system.file("ext-data",
                          "config.yaml",
                          package = "aftables")),
    sep = "\n")
cat("\n```\n")

## ----config example 1, eval = FALSE-------------------------------------------
# my_wb <- generate_workbook(aftable = demo_aftable)

## ----config example 2, eval = FALSE-------------------------------------------
# my_wb <- generate_workbook(
#   aftable = demo_aftable,
#   config_name = "workbook1"
# )

## ----config example 3, eval = FALSE-------------------------------------------
# dir.create("configs")
# 
# create_config_yaml(
#   path = "configs",
#   open_config = FALSE
# )
# 
# my_wb <- generate_workbook(
#   aftable = demo_aftable,
#   config_path = "configs/config.yaml"
# )

## ----config example 4, eval = FALSE-------------------------------------------
# dir.create("configs")
# 
# create_config_yaml(
#   path = "configs",
#   open_config = FALSE
# )
# 
# my_wb <- generate_workbook(
#   aftable = demo_aftable,
#   config_path = "configs/config.yaml",
#   config_name = "workbook1"
# )

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)

## ----install from CRAN, eval=FALSE--------------------------------------------
# install.packages("aftables")

## ----tables-cover-list--------------------------------------------------------
cover_list <- list(
  "Section 1" = c("First row of Section 1.", "Second row of Section 1."),
  "Section 2" = "The only row of Section 2.",
  "Section 3" = c(
    "[Website](https://best-practice-and-impact.github.io/aftables/)",
    "[Email address](mailto:fake.address@aftables.com)"
  )
)

## ----tables-contents----------------------------------------------------------
contents_df <- data.frame(
  "Sheet name" = c("Notes", "Table_1", "Table_2"),
  "Sheet title" = c(
    "Notes used in this workbook",
    "First Example Sheet",
    "Second Example Sheet"
  ),
  check.names = FALSE
)

## ----tables-notes-------------------------------------------------------------
notes_df <- data.frame(
  "Note number" = paste0("[note ", 1:3, "]"),
  "Note text" = c("First note.", "Second note.", "Third note."),
  check.names = FALSE
)

## ----tables-meta-expand-------------------------------------------------------
cover_list
contents_df
notes_df

## ----stats-df-1---------------------------------------------------------------
table_1_df <- data.frame(
  Category = LETTERS[1:10],
  "Numeric [note 1]" = 1:10,
  "Numeric suppressed" = c(1:4, "[c]", 6:9, "[x]"),
  "Numeric thousands" = abs(round(rnorm(10), 4) * 1e5),
  "Numeric decimal" = abs(round(rnorm(10), 5)),
  "Long name that means that the column width needs to be widened" = 1:10,
  Notes = c("[note 1]", rep(NA_character_, 4), "[note 2]", rep(NA_character_, 4)),
  check.names = FALSE
)

## ----stats-df-2---------------------------------------------------------------
table_2_df <- data.frame(Category = LETTERS[1:10], Numeric = 1:10)

## ----table-stat-expand--------------------------------------------------------
table_1_df
table_2_df

## ----new-aftable--------------------------------------------------------------
my_aftable <- aftables::create_aftable(
  tab_titles = c("Cover", "Contents", "Notes", "Table 1", "Table_2"),
  sheet_types = c("cover", "contents", "notes", "tables", "tables"),
  sheet_titles = c(
    "The 'aftables' Demo Workbook",
    "Table of contents",
    "Notes",
    "Table 1: First Example Sheet",
    "Table 2: Second Example Sheet"
  ),
  blank_cells = c(
    rep(NA_character_, 3),
    "Blank cells indicate that there's no note in that row.",
    NA_character_
  ),
  custom_rows = list(
    NA_character_,
    NA_character_,
    "A custom row.",
    c(
      "First custom row [with a hyperlink.](https://best-practice-and-impact.github.io/aftables/)",
      "Second custom row."
    ),
    "A custom row."
  ),
  sources = c(
    rep(NA_character_, 3),
    "[The Source Material., 2024](https://best-practice-and-impact.github.io/aftables/)",
    "The Source Material, 2024."
  ),
  tables = list(cover_list, contents_df, notes_df, table_1_df, table_2_df)
)

## ----aftable-preview----------------------------------------------------------
my_aftable

## ----create-af-wb-------------------------------------------------------------
my_wb <- aftables::generate_workbook(my_aftable)

## ----wb-preview---------------------------------------------------------------
my_wb

## ----saveworkbook, eval=FALSE-------------------------------------------------
# openxlsx2::wb_save(wb = my_wb, file = "C://Documents//publication.xlsx")

#' Insert 'create_aftable' template
#'
#' Insert at the cursor a template for \code{\link{create_aftable}} from the
#' 'aftables' package, pre-filled with demo data.
#'
#' @return Empty list. Function is used for side effect.
#'
#' @export
at_template_aftable <- function() {
  rstudioapi::insertText(string_create_aftable())
}

#' Insert full 'aftables' template workflow
#'
#' Insert at the cursor (a) demo templates for cover, contents and notes
#' tables, and (b) a call to \code{\link{create_aftable}} pre-filled with
#' demo data.
#'
#' @return Empty list. Function is used for side effect.
#'
#' @export
at_template_workflow <- function() {
  rstudioapi::insertText(
    paste0(
      "# Prepare tables of information",
      "\n\n",
      string_tables(),
      "\n\n",
      "# Create new aftable",
      "\n\n",
      string_create_aftable(),
      "\n\n",
      "# Generate workbook from aftable",
      "\n\n",
      "my_wb <- aftables::generate_workbook(my_aftable)",
      "\n\n",
      "# Create output",
      "\n\n",
      "openxlsx2::wb_open(my_wb)  # open temp copy",
      "\n\n",
      'openxlsx2::wb_save(my_wb, "example.xlsx")  # change save location'
    )
  )
}

#' A String Containing Code to Prepare Tables for an 'aftables' Object
#' @noRd
string_tables <- function() {
  'cover_list <- list(
  "Section 1" = c("First row of Section 1.", "Second row of Section 1."),
  "Section 2" = "The only row of Section 2.",
  "Section 3" = c(
    "[Website](https://best-practice-and-impact.github.io/aftables/)",
    "[Email address](mailto:fake.address@aftables.com)"
  )
)

contents_df <- data.frame(
  "Sheet name" = c("Notes", "Table 1", "Table 2"),
  "Sheet title" = c(
    "Notes used in this workbook",
    "First Example Sheet",
    "Second Example Sheet"
  ),
  check.names = FALSE
)

notes_df <- data.frame(
  "Note number" = paste("[note ", 1:2, "]"),
  "Note text" = c("First note.", "Second note."),
  check.names = FALSE
)

table_1_df <- data.frame(
  Category = LETTERS[1:10],
  Numeric = 1:10,
  "Numeric suppressed" = c(1:4, "[c]", 6:9, "[x]"),
  "Numeric thousands" = abs(round(rnorm(10), 4) * 1e5),
  "Numeric decimal" = abs(round(rnorm(10), 5)),
  "Long name that means that the column width needs to be widened" = 1:10,
  Notes = c("[note 1]", rep(NA_character_, 4), "[note 2]", rep(NA_character_, 4)),
  check.names = FALSE
)

table_2_df <- data.frame(Category = LETTERS[1:10], Numeric = 1:10)'
}

#' A String Containing Code to Generate an 'aftables' Object
#' @noRd
string_create_aftable <- function() {
  'my_aftable <-
  aftables::create_aftable(
    tab_titles = c("Cover", "Contents", "Notes", "Table_1", "Table_2"),
    sheet_types = c("cover", "contents", "notes", "tables", "tables"),
    sheet_titles = c(
      "The \'aftables\' Demo Workbook",
      "Table of contents",
      "Notes",
      "Table 1: First Example Sheet",
      "Table 2: Second Example Sheet"
    ),
    blank_cells = c(
      rep(NA_character_, 3),
      "Blank cells indicate that there\'s no note in that row.",
      NA_character_
    ),
    custom_rows = list(
      NA_character_,
      "A custom row in the Contents sheet.",
      NA_character_,
      c(
        "First custom row for Table 1.",
        "A second custom row [with a hyperlink.](https://best-practice-and-impact.github.io/aftables/)"
      ),
      "A custom row for Table 2"
    ),
    sources = c(
      rep(NA_character_, 3),
      "[The Source Material, 2024](https://best-practice-and-impact.github.io/aftables/)",
      "The Source Material, 2024"
    ),
    tables = list(cover_list, contents_df, notes_df, table_1_df, table_2_df)
  )'
}

test_that("string_create_aftable skeleton is okay", {
  expect_snapshot_output(string_create_aftable())
})

test_that("string_tables_tibble skeleton is okay", {
  expect_snapshot_output(string_tables())
})

#' Generate A Workbook Object From An 'aftable'
#'
#' Populate an 'openxlsx2' wbWorkbook-class object with content from an
#' aftable-class object. In turn, the output can be passed to
#' \code{\link[openxlsx2]{wb_save}} from 'openxlsx2'
#'
#' @param aftable An aftable-class object created using
#'     \code{\link{create_aftable}} (or \code{\link{as_aftable}}), which
#'     contains the data and information needed to create a workbook.
#' @param author Optional character string to set the workbook author.
#' Default NULL.
#' @param title Optional character string to set the workbook title.
#' Default NULL.
#' @param keywords Optional character vector to set the workbook keywords.
#' Default NULL.
#' @param config_path Optional character string containing path to config file.
#' Defaults to config.yaml file located in working directory.
#' @param config_name Optional character string specifying which configuration
#'   to use from config file. Default NULL.
#'
#' @return An openxlsx2 wbWorkbook-class object.
#'
#' @details
#'
#' Analysis Function guidance advises workbooks should have the author, title,
#' keywords and language document properties completed. `aftables` provides
#' functionality to set the author, title and keywords properties.
#' See [Releasing statistics in spreadsheets](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/#section-15)
#' for more information including how to set the workbook language.
#'
#' A config file can be used to set workbook properties and formatting. See
#' \code{\link[aftables]{create_config_yaml}} for details of how to create a
#' config.yaml file. If author, title or keywords are provided in both the
#' `config.yaml` file and in the `generate_workbook` arguments, the values provided in the
#' function arguments are preferred to those provided in the `config.yaml` file.
#'
#' @examples
#' # Convert an aftable to an openxlsx2 wbWorkbook-class object
#' # Setting the minimum workbook properties as function arguments
#' \dontrun{
#' example_workbook <- generate_workbook(
#'   demo_aftable,
#'   author = "Example author",
#'   title = "example workbook",
#'   keywords = c("keyword1", "keyword2", "keyword3")
#' )}
#'
#' # Use openxlsx2::wb_get_properties to view properties that have been applied
#' \dontrun{
#' openxlsx2::wb_get_properties(example_workbook)}
#'
#' # Save the workbook with openxlsx2::wb_save
#' \dontrun{
#' openxlsx2::wb_save(example_workbook, "example_workbook.xlsx")}
#'
#' # Using config.yaml file to set workbook properties and edit text and cell
#' # formatting
#' \dontrun{
#' example_workbook2 <- generate_workbook(
#'   demo_aftable,
#'   config_path = system.file("ext-data", "config.yaml", package = "aftables"),
#'   config_name = "workbook1"
#'   )}
#'
#' # Use openxlsx2::wb_get_properties to view properties that have been applied
#' \dontrun{
#' openxlsx2::wb_get_properties(example_workbook2)}
#'
#' @export
generate_workbook <- function(aftable,
                              author = NULL,
                              title = NULL,
                              keywords = NULL,
                              config_path = "config.yaml",
                              config_name = NULL) {

  if (!is_aftable(aftable)) {
    stop("The object passed to argument 'content' must have class 'aftable'.")
  }

  if (!is.null(author) && !rlang::is_scalar_character(author)) {
    stop("author must be a character vector of length 1", call. = FALSE)
  }

  if (!is.null(title) && !rlang::is_scalar_character(title)) {
    stop("title must be a character vector of length 1", call. = FALSE)
  }

  if (!is.null(keywords) && !is.character(keywords)) {
    stop("keywords must be a character vector", call. = FALSE)
  }

  user_config <- list(
    workbook_properties = list(
      author = author,
      title = title,
      keywords = keywords
    ),
    workbook_format = list()
  )

  config <- .process_config(user_config, config_path, config_name)

  workbook_properties <- config$workbook_properties
  workbook_format <- config$workbook_format

  font_ref <- .style_font(workbook_format)

  # Create a table_name from tab_title (unique, no spaces, no punctuation)
  aftable[["table_name"]] <-
    gsub(" ", "_", tolower(trimws(aftable[["tab_title"]])))
  aftable[["table_name"]] <-
    gsub("(?!_)[[:punct:]]", "", aftable[["table_name"]], perl = TRUE)

  # Create workbook, set base style, set properties, add tabs, cover, contents (required for all workbooks)
  wb <- wb_workbook(theme = "Office 2007 - 2010 Theme")
  wb <- .set_workbook_properties(wb, workbook_properties)
  wb <- .style_workbook(wb, font_ref)
  wb <- .add_tabs(wb, aftable)
  wb <- .add_cover(wb, aftable, font_ref)
  wb <- .add_contents(wb, aftable, font_ref, workbook_format)

  # There won't always be a notes tab
  if (any(aftable$sheet_type %in% "notes")) {
    wb <- .add_notes(wb, aftable, font_ref, workbook_format)
  }

  # Iterable titles for tabs containing tables
  table_sheets <- aftable[aftable$sheet_type == "tables", ][["table_name"]]

  for (i in table_sheets) {
    wb <- .add_tables(wb, aftable, table_name = i, font_ref, workbook_format)
  }

  wb
}

test_that("workbook object is created", {
  expect_warning(
    wb <- generate_workbook(as_aftable(demo_df)),
    "Some of the recommended workbook properties are missing."
  )

  expect_s3_class(wb, class = c("wbWorkbook", "R6"))
  expect_identical(class(wb)[1], "wbWorkbook")
})

test_that("aftable is passed", {
  wb <- suppressWarnings(as_aftable(demo_df))

  expect_error(generate_workbook("wb"))
  expect_error(generate_workbook(1))
  expect_error(generate_workbook(list()))
  expect_error(generate_workbook(data.frame()))
})


test_that("Error if workbook properties arguments are not correct type", {

  expect_error(
    generate_workbook(demo_aftable, author = 1),
    "author must be a character vector of length 1"
  )

  expect_error(
    generate_workbook(demo_aftable, title = c("a", "b")),
    "title must be a character vector of length 1"
  )

  expect_error(
    generate_workbook(demo_aftable, keywords = 2),
    "keywords must be a character vector"
  )

})



test_that(".stop_bad_input works as intended", {
  wb <- openxlsx2::wb_workbook()

  test_aftable <- as_aftable(demo_df)

  expect_error(.stop_bad_input("wb", test_aftable, "cover"),
               "'wb' must be an openxlsx2 wbWorkbook-class object.")
  expect_error(.stop_bad_input(wb, test_aftable, 1),
               "'table_name' must be a string of length 1")
})

test_that("hyperlinks are generated on the cover page", {
  # demo dataset has two hyperlinks on the cover
  expect_warning(
    wb <- generate_workbook(as_aftable(demo_df)),
    "Some of the recommended workbook properties are missing."
  )

  expect_equal(sum(grepl("HYPERLINK", wb$worksheets[[1]]$sheet_data$cc$f)), 2)
})

# Test .process_config ---------------------------------------------------------

test_that("error if config file doesn't exist", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      author = NULL,
      title = NULL,
      keywords = NULL,
      config_path = testthat::test_path("configs/DOES_NOT_EXIST.yaml"),
      config_name = NULL
    ),
    "DOES_NOT_EXIST.yaml does not exist"
  )

})

test_that("error if config doesn't have aftables key", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      author = NULL,
      title = NULL,
      keywords = NULL,
      config_path = testthat::test_path("configs/test_missing_aftable_config.yaml"),
      config_name = NULL
    ),
    "does not contain an aftables key"
  )

})

test_that("error if default is missing and not using custom key", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      author = NULL,
      title = NULL,
      keywords = NULL,
      config_path = testthat::test_path("configs/test_missing_default_config.yaml"),
      config_name = NULL
    ),
    "does not contain a default aftables configuration and a custom key is not being used"
  )

})

test_that("error if default config not a named list", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_default.yaml")
    ),
    "Default configuration key must be a named list. It can only contain keys `workbook_properties` and `workbook_format`."
  )

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_default2.yaml")
    ),
    "Default configuration key must be a named list. It can only contain keys `workbook_properties` and `workbook_format`."
  )

})

test_that("error if default workbook properties is not a named list", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_default3.yaml")
    ),
    "Default configuration workbook_properties must be a named list"
  )

})

test_that("error if default workbook format is not a named list", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_default4.yaml")
    ),
    "Default configuration workbook_format must be a named list"
  )

})

test_that("error if aftables cannot find custom config requested by user", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_config.yaml"),
      config_name = "MISSING_custom"
    ),
    "does not contain key `MISSING_custom`"
  )

})

test_that("error if custom config not a named list", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_custom.yaml"),
      config_name = "custom"
    ),
    "Configuration key custom must be a named list. It can only contain keys `workbook_properties` and `workbook_format`."
  )

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_custom2.yaml"),
      config_name = "custom"
    ),
    "Configuration key custom must be a named list. It can only contain keys `workbook_properties` and `workbook_format`."
  )

})

test_that("error if custom workbook properties is not a named list", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_custom3.yaml"),
      config_name = "custom"
    ),
    "Custom configuration workbook_properties must be a named list"
  )

})

test_that("error if custom workbook format is not a named list", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_custom4.yaml"),
      config_name = "custom"
    ),
    "Custom configuration workbook_format must be a named list"
  )

})

test_that(".process_config correctly combines config options", {

  # User + default + custom config ---------------------------------------------

  user_config <- list(
    workbook_properties = list(
      author = "user author",
      title = NULL,
      keywords = c("user", "keywords")
    ),
    workbook_format = list()
  )

  config <- .process_config(
    user_config = user_config,
    config_path = testthat::test_path("configs/test_valid_config.yaml"),
    config_name = "custom"
  )

  expected_config <- list(
    workbook_properties = list(
      author = "user author",
      title = "default config title",
      keywords = c("user", "keywords"),
      subject = "default config subject",
      category = "default config category",
      comments = "default config comments"
    ),
    workbook_format = list(
      base_font_name = "default font",
      base_font_size = 1,
      table_header_size = 1,
      sheet_subheading_size = 26,
      sheet_heading_size = 50,
      cellwidth_default = 1,
      cellwidth_wider = 1,
      nchar_break = 1
    )
  )

  expect_equal(config, expected_config)


  # User + default config ------------------------------------------------------

  config_default <- .process_config(
    user_config = user_config,
    config_path = testthat::test_path("configs/test_valid_config.yaml"),
    config_name = NULL
  )

  expected_config_default <- list(
    workbook_properties = list(
      author = "user author",
      title = "default config title",
      keywords = c("user", "keywords"),
      subject = "default config subject",
      category = "default config category",
      comments = "default config comments"
    ),
    workbook_format = list(
      base_font_name = "default font",
      base_font_size = 1,
      table_header_size = 1,
      sheet_subheading_size = 1,
      sheet_heading_size = 1,
      cellwidth_default = 1,
      cellwidth_wider = 1,
      nchar_break = 1
    )
  )

  expect_equal(config_default, expected_config_default)


  # Custom config only ---------------------------------------------------------

  config_custom <- .process_config(
    user_config = list(workbook_properties = list(), workbook_format = list()),
    config_path = testthat::test_path("configs/test_valid_config2.yaml"),
    config_name = "custom"
  )

  expected_config_custom <- list(
    workbook_properties = list(
      author = "custom config author",
      title = "custom config title",
      keywords = c("customconfig1", "customconfig2")
    ),
    workbook_format = list(
      sheet_subheading_size = 26,
      sheet_heading_size = 50
    )
  )

  expect_equal(config_custom, expected_config_custom)


  # user only ------------------------------------------------------------------
  # this test needs title to be set to avoid recommended properties warning
  user_config <- list(
    workbook_properties = list(
      author = "user author",
      title = "user title",
      keywords = c("user", "keywords")
    ),
    workbook_format = list()
  )

  config_user <- .process_config(
    user_config = user_config,
    config_path = NULL,
    config_name = NULL
  )

  expected_config_user <- list(
    workbook_properties = list(
      author =  "user author",
      title =  "user title",
      keywords = c("user", "keywords")
    ),
    workbook_format = list()
  )

  expect_equal(config_user, expected_config_user)

})

test_that("warnings for config file with identical values to example config", {

  expect_warning(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_same_config.yaml")
    ),
    "Your config file contains values identical to the aftables example config. Please check your config file."
  )

})


# Test .validate_config --------------------------------------------------------

test_that("error when config entries have invalid names or in wrong place", {

  expect_warning(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_config.yaml")
    ),
    "Some entries in your config file could not be processed."
  )

})

test_that("error when entries apart from keywords have more than 1 value", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_lengths_config.yaml"),
      config_name = "wrong-lengths"
    ),
    "Config entries must contain only one value apart from keywords. Please check your config file."
  )
})

test_that("error when values in config.yaml are wrong datatype (character/numeric/list)", {

  expect_error(
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_wrong_datatypes_config.yaml"),
      config_name = "wrong-datatypes"
    ),
    "Please review the following invalid config entries"
  )

})

test_that("Warn users when they haven't provided AF recommended workbook properties", {

  # Warn if config from file is missing recommended entries
  expect_warning(
    generate_workbook(
      as_aftable(demo_df),
      author = "aftables test",
      config_path = testthat::test_path("configs/test_missing_arguments_config.yaml")
    ),
    "Some of the recommended workbook properties are missing. Analysis Function guidance recommends completing the author, title and keywords fields."
  )

  # Warn if user does not provide all recommended arguments
  expect_warning(
    generate_workbook(
      as_aftable(demo_df),
      author = "aftables test",
      config_path = NULL
    ),
    "Some of the recommended workbook properties are missing. Analysis Function guidance recommends completing the author, title and keywords fields."
  )

})

test_that("no config is applied without config.yaml or function arguments", {
  expect_warning(
    wb <-
      generate_workbook(
        as_aftable(demo_df),
        config_path = NULL
      ),
    "Some of the recommended workbook properties are missing."
  )

  wb_properties <- openxlsx2::wb_get_properties(wb)

  # only 4 properties should be set (by openxlsx2 when wbWorkbook object created)
  expect_equal(
    names(wb_properties),
    c("creator", "datetime_created", "datetime_modified", "modifier")
  )

})

test_that("minimum properties are applied correctly via arguments", {

  wb <- generate_workbook(
    as_aftable(demo_df),
    author = "aftables author",
    title = "aftables test workbook arguments",
    keywords =  c("aftablesexample", "aftablesdemonstration",
                  "aftablesconfig.yaml"),
    config_path = NULL
  )

  wb_properties <- openxlsx2::wb_get_properties(wb)

  # only 6 properties should be set
  expect_equal(
    names(wb_properties),
    c("creator",
      "datetime_created",
      "datetime_modified",
      "modifier",
      "title",
      "keywords")
  )

  # minimum properties
  expect_equal(wb_properties["creator"],
               c("creator" = "aftables author"))
  expect_equal(wb_properties["title"],
               c("title" = "aftables test workbook arguments"))
  expect_equal(wb_properties["keywords"],
               c("keywords" = "aftablesexample, aftablesdemonstration, aftablesconfig.yaml"))

  # if modifier is blank it is populated with value for author
  expect_equal(wb_properties["modifier"],
               c("modifier" = "aftables author"))

})

test_that("properties from config.yaml are ignored when properties arguments are set", {

  wb <- generate_workbook(
    as_aftable(demo_df),
    author = NULL,
    title = "title fun argument",
    keywords =  c("keywords", "fun", "argument"),
    config_path = testthat::test_path("configs/test_custom_config.yaml"),
    config_name = "custom-config"
  )

  wb_properties <- openxlsx2::wb_get_properties(wb)

  # Properties from function arguments
  expect_equal(wb_properties["title"],
               c("title" = "title fun argument"))
  expect_equal(wb_properties["keywords"],
               c("keywords" = "keywords, fun, argument"))

  # Properties from custom config
  expect_equal(wb_properties["creator"],
               c("creator" = "Author custom config"))

  # Properties from default config
  expect_equal(wb_properties["subject"],
               c("subject" = "Subject default config test"))

})

test_that("Cell formats are set properly when no config set", {

  wb <- generate_workbook(
    as_aftable(demo_df),
    config_path = NULL,
    author = "test author",
    title = "test tile",
    keywords = "test keywords"
  )

  wb_fonts <- wb$styles_mgr$font

  wb_xf <- wb$styles_mgr$xf

  # Sheet heading
  xf_font_id <- openxlsx2::wb_get_cell_style(wb, "Table_1", "A1") |> as.integer()

  font_id <- (wb_xf |> filter(id == xf_font_id))$name

  font_id <- sub(".*fontId=\"([0-9]+)\".*", "\\1", font_id) |> as.numeric()

  cell_format <- (wb_fonts |> filter(id == font_id))$name

  expect_true(stringr::str_detect(cell_format, "<b val=\"1\"/>")) # bold
  expect_true(stringr::str_detect(cell_format, "<name val=\"Arial\"/>")) # Arial
  expect_true(stringr::str_detect(cell_format, "<sz val=\"16\"/>")) # size 16

  # Sheet subheading
  xf_font_id <- openxlsx2::wb_get_cell_style(wb, "Cover", "A2") |> as.integer()

  font_id <- (wb_xf |> filter(id == xf_font_id))$name

  font_id <- sub(".*fontId=\"([0-9]+)\".*", "\\1", font_id) |> as.numeric()

  cell_format <- (wb_fonts |> filter(id == font_id))$name

  expect_true(stringr::str_detect(cell_format, "<b val=\"1\"/>")) # bold
  expect_true(stringr::str_detect(cell_format, "<name val=\"Arial\"/>")) # Arial
  expect_true(stringr::str_detect(cell_format, "<sz val=\"14\"/>")) # size 14

  # table header
  xf_font_id <- openxlsx2::wb_get_cell_style(wb, "Table_1", "A8") |> as.integer()

  font_id <- (wb_xf |> filter(id == xf_font_id))$name

  font_id <- sub(".*fontId=\"([0-9]+)\".*", "\\1", font_id) |> as.numeric()

  cell_format <- (wb_fonts |> filter(id == font_id))$name

  expect_true(stringr::str_detect(cell_format, "<b val=\"1\"/>")) # bold
  expect_true(stringr::str_detect(cell_format, "<name val=\"Arial\"/>")) # Arial
  expect_true(stringr::str_detect(cell_format, "<sz val=\"12\"/>")) # size 12

  # default font Arial size 12
  # get properties with wb_get_base_font
  base_font <- openxlsx2::wb_get_base_font(wb)

  expect_equal(base_font$size$val, "12")

  expect_equal(base_font$name$val, "Arial")

})

test_that("Cell formats are set properly from default and custom configs combined", {

  wb <-
    generate_workbook(
      as_aftable(demo_df),
      config_path = testthat::test_path("configs/test_mixed_font_config.yaml"),
      config_name = "workbook1"
    )

  wb_fonts <- wb$styles_mgr$font

  wb_xf <- wb$styles_mgr$xf

  # Sheet heading
  xf_font_id <- openxlsx2::wb_get_cell_style(wb, "Table_1", "A1") |> as.integer()

  font_id <- (wb_xf |> filter(id == xf_font_id))$name

  font_id <- sub(".*fontId=\"([0-9]+)\".*", "\\1", font_id) |> as.numeric()

  cell_format <- (wb_fonts |> filter(id == font_id))$name

  expect_true(stringr::str_detect(cell_format, "<b val=\"1\"/>")) # bold
  expect_true(stringr::str_detect(cell_format, "<name val=\"Calibri\"/>")) # custom Calibri
  expect_true(stringr::str_detect(cell_format, "<sz val=\"20\"/>")) # custom size 20

  # Sheet subheading
  xf_font_id <- openxlsx2::wb_get_cell_style(wb, "Cover", "A2") |> as.integer()

  font_id <- (wb_xf |> filter(id == xf_font_id))$name

  font_id <- sub(".*fontId=\"([0-9]+)\".*", "\\1", font_id) |> as.numeric()

  cell_format <- (wb_fonts |> filter(id == font_id))$name

  expect_true(stringr::str_detect(cell_format, "<b val=\"1\"/>")) # bold
  expect_true(stringr::str_detect(cell_format, "<name val=\"Calibri\"/>")) # custom Calibri
  expect_true(stringr::str_detect(cell_format, "<sz val=\"18\"/>")) # custom size 18

  # table header
  xf_font_id <- openxlsx2::wb_get_cell_style(wb, "Table_1", "A8") |> as.integer()

  font_id <- (wb_xf |> filter(id == xf_font_id))$name

  font_id <- sub(".*fontId=\"([0-9]+)\".*", "\\1", font_id) |> as.numeric()

  cell_format <- (wb_fonts |> filter(id == font_id))$name

  expect_true(stringr::str_detect(cell_format, "<b val=\"1\"/>")) # bold
  expect_true(stringr::str_detect(cell_format, "<name val=\"Calibri\"/>")) # custom Calibri
  expect_true(stringr::str_detect(cell_format, "<sz val=\"16\"/>")) # custom size 16

  # default font Calibri size 8
  # get properties with wb_get_base_font
  base_font <- openxlsx2::wb_get_base_font(wb)

  expect_equal(base_font$size$val, "8")

  expect_equal(base_font$name$val, "Calibri")

})

test_that("Default column widths are set", {

  wb <- generate_workbook(
    as_aftable(demo_df),
    config_path = testthat::test_path("configs/test_widths_config.yaml")
  )

  # column widths of worksheet Table_1 are standard for columns 1 to 5 and 7
  # columns 6 is a wide column
  col_widths <- wb$worksheets[[4]]$cols_attr

  expect_true(stringr::str_detect(col_widths[1], "min=\"1\" max=\"5\"") &&
                stringr::str_detect(col_widths[1], "width=\"16.555\""))

  expect_true(stringr::str_detect(col_widths[2], "min=\"6\" max=\"6\"") &&
                stringr::str_detect(col_widths[2], "width=\"32.555\""))

  expect_true(stringr::str_detect(col_widths[3], "min=\"7\" max=\"7\"") &&
                stringr::str_detect(col_widths[3], "width=\"16.555\""))

})

test_that("Custom column widths are set from default and custom configs combined", {

  wb <- generate_workbook(
    as_aftable(demo_df),
    config_path = testthat::test_path("configs/test_custom_widths_config.yaml"),
    config_name = "workbook1"
  )

  # column widths of worksheet Table_1 are standard for columns 1 to 5 and 7
  # columns 6 is a wide column
  col_widths <- wb$worksheets[[4]]$cols_attr

  expect_true(stringr::str_detect(col_widths[1], "min=\"1\" max=\"5\"") &&
                stringr::str_detect(col_widths[1], "width=\"10.555\""))

  expect_true(stringr::str_detect(col_widths[2], "min=\"6\" max=\"6\"") &&
                stringr::str_detect(col_widths[2], "width=\"14.555\""))

  expect_true(stringr::str_detect(col_widths[3], "min=\"7\" max=\"7\"") &&
                stringr::str_detect(col_widths[3], "width=\"10.555\""))

})


test_that("Setting nchar_break different to default forces column widths to change", {

  wb <- generate_workbook(
    as_aftable(demo_df),
    config_path = testthat::test_path("configs/test_nchar_break_config.yaml")
  )

  # nchar_break is set to 8
  # Column headers for columns 2 to 6 are greater than are 8 characters
  # so they are  widened
  # columns 1 and 7 are less than or equal to 8 characters so are not widened
  col_widths <- wb$worksheets[[4]]$cols_attr

  expect_true(stringr::str_detect(col_widths[1], "min=\"1\" max=\"1\"") &&
                stringr::str_detect(col_widths[1], "width=\"16.555\""))

  expect_true(stringr::str_detect(col_widths[2], "min=\"2\" max=\"6\"") &&
                stringr::str_detect(col_widths[2], "width=\"32.555\""))

  expect_true(stringr::str_detect(col_widths[3], "min=\"7\" max=\"7\"") &&
                stringr::str_detect(col_widths[3], "width=\"16.555\""))

})

#' Convert a List to A Sentence Form
#' Vectors of 1, 2 and 3 letters become 'A', 'A and B', 'A, B and C'.
#' @noRd
.vector_to_sentence <- function(vector) {
  if (length(vector) > 1) {
    last <- vector[length(vector)]
    not_last <- vector[-length(vector)]
    sentence <- paste(paste(not_last, collapse = ", "), "and", last)
    return(sentence)
  }

  vector
}

#' Create an aftables config.yaml
#'
#' Copy the example config.yaml file included with aftables to a directory of
#' the user's choice, and optionally open the file for editing. The config.yaml
#' file can be passed to the aftables function
#' \code{\link[aftables]{generate_workbook}}.
#'
#' @param path optional character string containing directory to copy the
#'   config.yaml file. Defaults to current working directory.
#' @param open_config optional logical whether to open the copy of config.yaml
#'   for editing in the current R session.
#'
#' @details If there is an existing config.yaml file in the destination
#' directory this function will not overwrite it.
#'
#' Contents of example config.yaml:
#'
#' ```{r results='asis', echo = FALSE}
#' cat('```xml\n')
#' cat(readLines(system.file("ext-data",
#'                           "config.yaml",
#'                           package = "aftables")),
#'     sep = "\n")
#' cat('\n```\n')
#' ```
#'
#' All configurations must be placed below an `aftables` key. `aftables` should
#' be followed by a `default` key and/or custom keys (e.g. `workbook1`).
#'
#' The `default` key settings will be read by all calls to
#' \code{\link[aftables]{generate_workbook}} which use this config.yml. This
#' allows you to share settings when generating multiple workbooks in one
#' script.
#'
#' Custom key settings (e.g. `workbook1`) will only be used by
#' \code{\link[aftables]{generate_workbook}} when the key is provided as the
#' `config_name` argument. This allows you to specify settings for a specific
#' workbook. Custom key settings will be preferred over the `default` settings.
#'
#' Keys below workbook_properties will appear in the Excel workbook when it is
#' saved using \code{\link[openxlsx2]{wb_save}} from openxlsx2. They can be
#' found in the file properties or the workbook information.
#'
#' Keys below `workbook_format` will be applied to the contents of the workbook.
#' The values of `base_font_name` and `base_font_size` define the default font
#' name and size used by the workbook. All text not formatted as a table header,
#' sheet subheading or sheet heading will use the default settings. Font sizes
#' of sheet headings, sheet subheadings, and table header rows will use the
#' values of `sheet_heading_size`, `sheet_subheading_size` and
#' `table_header_size` respectively, and they will additionally be formatted as
#' bold.
#'
#' The values of `cellwidth_default`, `cellwidth_wider` and `nchar_break` are
#' used to define column widths. The units of `cellwidth_default` and
#' `cellwidth_wider` are the column width values used by Excel. All columns
#' widths are set by default to use the `cellwidth_default` value. If the number
#' of characters in a column header or the contents of a column exceeds the
#' value of `nchar_break` aftables will set the column width to the value of
#' `cellwidth_wider`. Users can avoid text wrapping in columns or column headers
#' by setting the value of `nchar_break` based on their data or the content of
#' their column headers.
#'
#' Not all workbook configuration options need to be set. Required settings are
#' documented in \code{\link[aftables]{generate_workbook}}.
#'
#' @examples
#' # Use default arguments to copy `aftables` `config.yaml` file to the current
#' # working directory without opening the copied file for editing:
#'
#' \dontrun{
#' create_config_yaml(open_config = FALSE)}
#'
#' @export
create_config_yaml <- function(path = getwd(),
                               open_config = rlang::is_interactive()) {

  if (!file.exists(path)) {
    stop(
      paste0("The directory `", path, "` does not exist."),
      call. = FALSE
    )
  }

  if (file.exists(paste0(path, "/config.yaml"))) {
    stop(
      paste0("`", path, "/config.yaml` already exists."),
      call. = FALSE
    )
  }

  copy <- file.copy(
    from = system.file("ext-data", "config.yaml", package = "aftables"),
    to = path,
    overwrite = FALSE,
    copy.mode = FALSE
  )

  if (!copy) {
    stop("Error copying config.yaml",
         call. = FALSE)
  }

  if (path == getwd()) {
    warning(
      paste0("config.yaml copied to working directory. The default options for generate_workbook will use this file."),
      call. = FALSE
    )
  } else {
    message(
      paste0("config.yaml copied to ", path, " folder.")
    )
  }

  if (open_config) {
    utils::file.edit(paste0(path, "/config.yaml"))
  }

  invisible(NULL)

}


# Process config arguments
.process_config <- function(user_config, config_path, config_name) {

  # Get settings from config file if exists ------------------------------------

  if (!is.null(config_path) && config_path != "config.yaml" &&
        !file.exists(config_path)) {
    stop("Config file ", config_path, " does not exist", call. = FALSE)
  }

  if ((is.null(config_path) || !file.exists(config_path)) &&
        !is.null(config_name)) {
    stop(
      "config_name has been set but config file does not exist",
      call. = FALSE
    )
  }

  if (!is.null(config_path) && file.exists(config_path)) {

    config_file <- yaml::read_yaml(config_path)

    if (config_path != "config.yaml" &&
          !pluck_exists(config_file, "aftables")) {
      stop(
        "Config file ", config_path, " does not contain an aftables key",
        call. = FALSE
      )
    }

    if (config_path == "config.yaml" &&
          !pluck_exists(config_file, "aftables")) {
      warning(
        "Config file ", config_path,
        " does not contain an aftables key and will therefore be ignored",
        call. = FALSE
      )
    }


    # Get default config settings ---------

    if (is.null(config_name) &&
          pluck_exists(config_file, "aftables") &&
          !pluck_exists(config_file, "aftables", "default")) {
      stop(
        "Config file ", config_path,
        " does not contain a default aftables configuration and a custom key is not being used",
        call. = FALSE
      )
    }

    if (pluck_exists(config_file, "aftables", "default")) {

      default_config <- pluck(config_file, "aftables", "default")

      if (!is.list(default_config) ||
        length(default_config) > 2 ||
        is.null(names(default_config)) ||
        !all(
          names(default_config) %in% c("workbook_properties", "workbook_format")
        )) {
        stop(
          "Default configuration key must be a named list. ",
          "It can only contain keys `workbook_properties` and `workbook_format`.",
          call. = FALSE
        )
      }

      if (pluck_exists(default_config, "workbook_properties") &&
            (!is.list(default_config$workbook_properties) ||
               is.null(names(default_config$workbook_properties)))) {
        stop(
          "Default configuration workbook_properties must be a named list",
          call. = FALSE
        )
      }

      if (pluck_exists(default_config, "workbook_format") &&
            (!is.list(default_config$workbook_format) ||
               is.null(names(default_config$workbook_format)))) {
        stop(
          "Default configuration workbook_format must be a named list",
          call. = FALSE
        )
      }

    } else {
      # No default config
      default_config <- list()
    }



    # Get custom config settings ---------

    if (!is.null(config_name)) {

      if (!pluck_exists(config_file, "aftables", config_name)) {
        stop(
          "Config file ", config_path, " does not contain key `",
          config_name, "`",
          call. = FALSE
        )
      }

      custom_config <- pluck(
        config_file,
        "aftables", config_name,
        .default = list()
      )

      if (!is.list(custom_config)  ||
        length(custom_config) > 2 ||
        is.null(names(custom_config)) ||
        !all(
          names(custom_config) %in% c("workbook_properties", "workbook_format")
        )) {
        stop(
          "Configuration key ", config_name, " must be a named list. ",
          "It can only contain keys `workbook_properties` and `workbook_format`.",
          call. = FALSE
        )
      }

      if (!is.null(custom_config$workbook_properties) &&
            (!is.list(custom_config$workbook_properties) ||
               is.null(names(custom_config$workbook_properties)))) {
        stop(
          "Custom configuration workbook_properties must be a named list",
          call. = FALSE
        )
      }

      if (!is.null(custom_config$workbook_format) &&
            (!is.list(custom_config$workbook_format) ||
               is.null(names(custom_config$workbook_format)))) {
        stop(
          "Custom configuration workbook_format must be a named list",
          call. = FALSE
        )
      }

    } else {
      # No custom config
      custom_config <- list()
    }


  } else {
    # No config file
    default_config <- list()
    custom_config <- list()
  }


  # Combine config options -----------------------------------------------------
  # Remove null options in user config - these are default function argument
  user_config <- map_depth(user_config, 1, compact)

  # Combine config settings, user config has highest priority, default config lowest
  config <-  list_modify(default_config, !!!custom_config)
  config <-  list_modify(config, !!!user_config)

  # Warning if the config workbook_properties have not been
  # changed from the internal config.yaml defaults
  if (length(config$workbook_properties) > 0 &&
      any(unlist(config$workbook_properties,
                 use.names = FALSE) %in%
          c("aftables",
            "aftables example workbook",
            "aftables workbook 2",
            "aftables1",
            "aftables2",
            "aftables3",
            "aftables example subject",
            "aftables example category",
            "aftables workbook 1 category",
            "aftablesworkbook1",
            "aftables workbook 2 category",
            "aftablesworkbook2",
            "aftables example comments")
      )
  ) {
    warning(
      paste0(
        "Your config file contains values identical to the aftables example ",
        "config. Please check your config file."
      ),
      call. = FALSE
    )
  }

  # Validate the final config --------------------------------------------------
  .validate_config(config)

  # Warning if recommended properties author, title or keywords are
  # missing from config
  if (!all(c(pluck_exists(config, "workbook_properties", "author"),
             pluck_exists(config, "workbook_properties", "keywords"),
             pluck_exists(config, "workbook_properties", "title")))) {
    warning(
      paste0("Some of the recommended workbook properties are missing. ",
             "Analysis Function guidance recommends completing the ",
             "author, title and keywords fields."),
      call. = FALSE
    )
  }

  config
}

# check config field datatypes
.validate_config <- function(config) {

  # The config may contain any of these entries and they should be these
  # datatypes. If the config contains any extra entries they won't be used by
  # functions. Functions will handle non-existent or empty entries as NULL.

  correct_datatypes <-
    tibble::tibble(
      correct_parent = c(
        rep("workbook_properties", 6),
        rep("workbook_format", 8)
      ),
      entry = c(
        "author",
        "category",
        "comments",
        "keywords",
        "subject",
        "title",
        "base_font_name",
        "base_font_size",
        "cellwidth_default",
        "cellwidth_wider",
        "nchar_break",
        "sheet_heading_size",
        "sheet_subheading_size",
        "table_header_size"
      ),
      datatype = c(
        rep("character", 7),
        rep("integer", 7)
      )
    )

  # Turn config list into dataframe
  config_df <-
    tibble::tibble(
      # create vector of workbook_properties and workbook_format
      # with length equal to number of entries under each
      parent = rep(map_depth(config, 0, names) |>
                     unlist(use.names = FALSE) %||% character(0),
                   map_depth(config, 1, length) |>
                     unlist(use.names = FALSE) %||% 0),
      # get names of each entry under workbook_properties and workbook_format
      entry =  map_depth(config, 1, names) |>
        unlist(use.names = FALSE) %||% character(0),
      # Find the datatype of each entry under workbook_properties and
      # workbook_format
      # Replace NULL with character(0) so column is always created
      config_datatype = map_depth(config, 2, typeof) |>
        unlist(use.names = FALSE) %||% character(0),
      # Find the length of each entry under workbook_properties and
      # workbook_format
      # Replace NULL with character(0) so column is always created
      config_length = map_depth(config, 2, length) |>
        unlist(use.names = FALSE) %||% character(0)
    )

  # Warning if entries in user config are not recognised
  # either wrong name or under the wrong entry (workbook_properties/workbook_format)
  unrecognised_entries <-
    dplyr::anti_join(
      config_df,
      correct_datatypes,
      by = dplyr::join_by("parent" == "correct_parent", "entry")
    ) |>
    mutate(message = paste0(.data$parent, ":", .data$entry))

  if (nrow(unrecognised_entries) > 0) {
    warning(
      paste0(
        "Some entries in your config file could not be processed.\n",
        "Entries with invalid names or in the wrong location will ",
        "not be added to your workbook.\n",
        "Please check the names and locations of these entries ",
        "in your config file:\n",
        paste0(unrecognised_entries$message, collapse = "\n")
      ),
      call. = FALSE
    )
  }

  # Combine the configs
  combined_configs <-
    dplyr::inner_join(
      correct_datatypes,
      config_df,
      by = dplyr::join_by("correct_parent" == "parent", "entry")
    )

  # Keywords can be any length, all other entries should be 1
  config_lengths <-
    combined_configs |>
    filter(.data$entry != "keywords" & .data$config_length > 1)

  # Error if any entries contain more than 1 value apart from keywords
  if (nrow(config_lengths) > 0) {
    stop(
      paste0(
        "Config entries must contain only one value apart from keywords. ",
        "Please check your config file for these entries: ",
        paste0(config_lengths$entry, collapse = ", "), "."
      ),
      call. = FALSE
    )
  }

  # Identify entries with incorrect datatypes and create error messages
  incorrect_datatypes <-
    combined_configs |>
    filter(.data$datatype != .data$config_datatype) |>
    mutate(
      across(
        c("datatype", "config_datatype"),
        \(x) str_replace(x, "character", "character string")
      ),
      across(
        c("datatype", "config_datatype"),
        \(x) str_replace(x, "integer", "integer value")
      ),
      error_message = paste0(.data$correct_parent, ":", .data$entry, " is ",
                             .data$config_datatype, ". It should be ",
                             .data$datatype, ".")
    )

  # Error if any invalid config datatypes
  if (nrow(incorrect_datatypes) > 0) {
    stop(
      c(
        "Please review the following invalid config entries:\n",
        paste0(incorrect_datatypes$error_message, collapse = "\n")
      ),
      call. = FALSE
    )
  }

  invisible(NULL)

}

#' Create An 'aftable' Object
#'
#' Create a new aftable-class object, which is a special data.frame that
#' contains all the information needed in your output spreadsheet. In turn, the
#' object created by this function can be used to populate an 'openxlsx2'
#' wbWorkbook-class object with the function \code{\link{generate_workbook}}.
#'
#' @param tab_titles Required character vector, one value per sheet. Each title
#'     will appear literally on each tab of the final spreadsheet output. Keep
#'     brief. Letters and numbers only; do not start with a number; use
#'     underscores for spaces. For example: 'Cover', 'Contents', 'Notes',
#'     'Table_1'. Will be corrected automatically unless there's an error.
#' @param sheet_types Required character vector, one value per sheet. Sheets
#'     that don't contain publication tables ('meta' sheets) should be of type
#'     'contents', 'cover' or 'notes'. Sheets that contain statistical tables of
#'     data are type 'tables'.
#' @param sheet_titles Required character vector, one value per sheet. The main
#'     title for each sheet, which will appear in cell A1 (top-left corner).
#' @param blank_cells Optional character vector, one value per sheet. A short
#'     sentence to explain the reason for any blank cells in the sheet. Supply
#'     as \code{NA_character_} if empty. Most likely to be used with sheet type
#'     'tables'.
#' @param custom_rows Optional list of character vectors. One list element per
#'     sheet, one character vector element per row of pre-table metadata. Supply
#'     a list element as \code{NA_character_} if empty. To be used with sheet
#'     type 'tables', but can also be used for sheet types 'contents' and
#'     'notes'.
#' @param sources Optional character vector, one value per sheet. The origin of
#'     the data for a given sheet. Supply as \code{NA_character_} if empty. To
#'     be used with sheet type 'tables'.
#' @param tables Required list of data.frames (though the cover sheet may be
#'     supplied as a list), one per sheet. See details.
#'
#' @details
#'
#' ## How to supply data to the 'tables' argument
#'
#' Formats for the elements collected as a list and passed to the 'tables'
#' argument, depending on the sheet type.
#'
#' \itemize{
#'     \item Sheet type 'cover': either (a) a list where each element name is a
#'         section header and each element's content is a character vector
#'         whose elements will make up separate rows of that section
#'         (recommended), or (b) a data.frame with one row per subsection, with
#'         one column for section titles and one column for corresponding for
#'         that section's body text. For example, you may have a section with
#'         the title 'Contact details' that contains an email address and
#'         telephone number. You can use line breaks (i.e. '\\n') to separate
#'         text into paragraphs.
#'     \item Sheet type 'contents': one row per sheet, two columns suggested at
#'         least (named 'Tab title' and 'Worksheet title').
#'     \item Sheet type 'notes': one row per note, two columns suggested (named
#'         'Note number', 'Note text'), where notes are in the form
#'         '\[note 1\]'.
#'     \item Sheet type 'tables': a tidy, rectangular data.frame containing the
#'         data to be published. It's the user's responsibility to add notes in
#'         the form '\[note 1\]' to column headers, or in a special 'Notes' row.
#' }
#'
#' ## How to supply hyperlinks
#'
#' You can provide text in Markdown link syntax (e.g.
#' '\[GOV.UK\](https://www.gov.uk)', adding 'mailto:' before an
#' email address) and the containing cell will be rendered as a hyperlink in the
#' output spreadsheet. Note that whole cells will become hyperlinks; there is no
#' support for selected words in a sentence to be rendered as a hyperlink.
#'
#' Hyperlinks can be supplied in the character strings to three arguments:
#'
#' \itemize{
#'     \item To the 'tables' argument for sheet type 'cover' only. It's
#'         recommended to supply the cover information as a list rather than a
#'         data.frame, which will allow you to make specific rows within a
#'         section (e.g. 'contact us') into hyperlinks.
#'     \item To the 'custom_rows' argument for sheets of type 'contents, 'notes'
#'        and 'tables'.
#'     \item To the 'source' argument for sheets of type 'table' only.
#' }
#'
#' @return An object with classes 'aftable', 'tbl' and 'data.frame'.
#'
#' @examples
#' # Prepare some demo tables of information
#'
#' set.seed(1066)
#'
#' cover_list <- list(
#'   "Section 1" = c("First row of Section 1.", "Second row of Section 1."),
#'   "Section 2" = "The only row of Section 2.",
#'   "Section 3" = c(
#'     "[Website](https://best-practice-and-impact.github.io/aftables/)",
#'     "[Email address](mailto:fake.address@aftables.com)"
#'   )
#' )
#'
#' contents_df <- data.frame(
#'   "Sheet name" = c("Notes", "Table_1", "Table_2"),
#'   "Sheet title" = c(
#'     "Notes used in this workbook",
#'     "First Example Sheet",
#'     "Second Example Sheet"
#'   ),
#'   check.names = FALSE
#' )
#'
#' notes_df <- data.frame(
#'   "Note number" = paste0("[note ", 1:2, "]"),
#'   "Note text" = c("First note.", "Second note."),
#'   check.names = FALSE
#' )
#'
#'
#' table_1_df <- data.frame(
#'   Category = LETTERS[1:10],
#'   "Numeric [note 1]" = 1:10,
#'   "Numeric suppressed" = c(1:4, "[c]", 6:9, "[x]"),
#'   "Numeric thousands" = abs(round(rnorm(10), 4) * 1e5),
#'   "Numeric decimal" = abs(round(rnorm(10), 5)),
#'   "Long name that means that the column width needs to be widened" = 1:10,
#'   Notes = c(
#'     "[note 1]", rep(NA_character_, 4), "[note 2]",
#'     rep(NA_character_, 4)
#'   ),
#'   check.names = FALSE
#' )
#'
#' table_2_df <- data.frame(Category = LETTERS[1:10], Numeric = 1:10)
#'
#' # Create 'aftables' object
#'
#' x <- aftables::create_aftable(
#'   tab_titles = c("Cover", "Contents", "Notes", "Table_1", "Table_2"),
#'   sheet_types = c("cover", "contents", "notes", "tables", "tables"),
#'   sheet_titles = c(
#'     "The 'aftables' Demo Workbook",
#'     "Table of contents",
#'     "Notes",
#'     "Table 1: First Example Sheet",
#'     "Table 2: Second Example Sheet"
#'   ),
#'   blank_cells = c(
#'     rep(NA_character_, 3),
#'     "Blank cells indicate that there's no note in that row.",
#'     NA_character_
#'   ),
#'   custom_rows = list(
#'     NA_character_,
#'     NA_character_,
#'     "A custom row.",
#'     c(
#'       paste0(
#'         "First custom row [with a hyperlink.]",
#'         "(https://best-practice-and-impact.github.io/aftables/)"
#'       ),
#'       "Second custom row."
#'     ),
#'     "A custom row."
#'   ),
#'   sources = c(
#'     rep(NA_character_, 3),
#'     paste0(
#'       "[The Source Material, 2024.]",
#'       "(https://best-practice-and-impact.github.io/aftables/)"
#'     ),
#'     "The Source Material, 2024."
#'   ),
#'   tables = list(cover_list, contents_df, notes_df, table_1_df, table_2_df)
#' )
#'
#' # Test that 'aftable' is one of the object's classes
#' is_aftable(x)
#'
#' # Look at the structure of the object
#' str(x, max.level = 2)
#'
#' @export
create_aftable <- function(tab_titles,
                           sheet_types = c("cover", "contents", "notes", "tables"),
                           sheet_titles,
                           blank_cells = NA_character_,
                           sources = NA_character_,
                           custom_rows = list(NA_character_),
                           tables) {

  x <- data.frame(
    tab_title = unlist(tab_titles),
    sheet_type = unlist(sheet_types),
    sheet_title = unlist(sheet_titles),
    blank_cells = unlist(blank_cells),
    source = unlist(sources),
    stringsAsFactors = FALSE # because default is TRUE prior to R v4
  )

  x[["custom_rows"]] <- custom_rows
  x[["table"]] <- tables

  as_aftable(x)
}

#' Coerce To An 'aftable' Object
#'
#' Functions to check if an object is an aftable, or coerce it if possible.
#'
#' @param x A data.frame object to coerce.
#'
#' @return \code{as_aftable} returns an object of class aftable if possible.
#'     \code{is_aftable} returns \code{TRUE} if the object has class
#'     aftable, otherwise \code{FALSE}.
#'
#' @examples
#' as_aftable(demo_df)
#' is_aftable(demo_aftable)
#'
#' @export
as_aftable <- function(x) {
  if (any(names(x) %in% "tab_title")) {
    .check_tab_titles(x[["tab_title"]])
    x[["tab_title"]] <- .clean_tab_titles(x[["tab_title"]])
  }

  if (any(names(x) %in% "blank_cells")) {
    x[["blank_cells"]] <- .append_period(x[["blank_cells"]])
  }

  class(x) <- c("aftable", "tbl", "data.frame")

  .validate_aftable(x)
  .warn_aftable(x)

  x
}

#' @rdname as_aftable
#' @export
is_aftable <- function(x) {
  inherits(x, "aftable")
}

#' Summarise An 'aftable' Object
#'
#' A concise result summary of an aftable-class object to see information
#' about the sheet content. Shows a numbered list of sheets with each tab title,
#' sheet type and table dimensions.
#'
#' @param object An aftable-class object for which to get a summary.
#' @param ... Other arguments to pass.
#'
#' @return `object` unaltered.
#'
#' @examples
#' # Print a concise summary of the aftable-class object
#' summary(demo_aftable)
#'
#' # Alternatively, look at the structure
#' str(demo_aftable, max.level = 2)
#'
#' @export
summary.aftable <- function(object, ...) {
  tables <- object[["table"]]

  table_dims <- vector("list", length = length(tables))

  for (i in seq_along(tables)) {
    if (inherits(tables[[i]], "list")) {
      list_length <- length(tables[[i]])
      list_lengths <- lengths(tables[[i]])

      table_dims[[i]] <- paste0(
        "list of length ", list_length,
        " (element lengths ", .vector_to_sentence(list_lengths), ")"
      )
    }

    if (is.data.frame(tables[[i]])) {
      table_dims[[i]] <-
        paste(paste(dim(tables[[i]]), collapse = " x "), "dataframe")
    }
  }

  summary_string <- paste0(
    "\n",
    paste0(
      paste0("  ", seq_along(tables), ") Tab '"),
      object[["tab_title"]],
      "' (sheet type '", object[["sheet_type"]], "') contains a ",
      unlist(table_dims),
      collapse = "\n"
    )
  )

  cat("# An aftable with", nrow(object), "sheets:", summary_string)

  invisible(object)
}


#' Provide A Succinct Summary Of An 'aftable' Object
#'
#' A brief text description of an aftable-class object.
#'
#' @param x An aftable-class object to summarise.
#' @param ... Other arguments to pass.
#'
#' @return Named character vector.
#'
#' @examples
#' # Print with description
#' print(demo_aftable)
#'
#' # Print description only (package 'tibble' must be installed)
#' tibble::tbl_sum(demo_aftable)
#'
#' @export
tbl_sum.aftable <- function(x, ...) {
  header <- sprintf(
    "%s x %s",
    formatC(nrow(x), big.mark = ","),
    formatC(ncol(x), big.mark = ",")
  )

  c("aftable" = header)
}

#' Set up a list of common paragraph styles
#' @noRd
.style_paragraph <- function() {
  list(
    lalign = "left",
    ralign = "right",
    wrap_text = TRUE
  )
}

#' Set up a list of common font styles
#' @noRd
.style_font <- function(workbook_format) {

  base_font_size <-
    if (!is.null(workbook_format$base_font_size))
      workbook_format$base_font_size else aftables_default_font$base_font_size

  table_header_size <-
    if (!is.null(workbook_format$table_header_size))
      workbook_format$table_header_size else aftables_default_font$table_header_size

  sheet_heading_size <-
    if (!is.null(workbook_format$sheet_heading_size))
      workbook_format$sheet_heading_size else aftables_default_font$sheet_heading_size

  sheet_subheading_size <-
    if (!is.null(workbook_format$sheet_subheading_size))
      workbook_format$sheet_subheading_size else aftables_default_font$sheet_subheading_size

  base_font_name <-
    if (!is.null(workbook_format$base_font_name))
      workbook_format$base_font_name else aftables_default_font$base_font_name

  list(
    bold =  1,
    base_font_size = base_font_size,
    table_header_size = table_header_size,
    sheet_heading_size = sheet_heading_size,
    sheet_subheading_size = sheet_subheading_size,
    name = base_font_name
  )
}

#' Apply Styles to the Whole Workbook
#' @param wb An 'openxlsx2' wbWorkbook object.
#' @noRd

.style_workbook <- function(wb, font_ref) {

  wb$set_base_font(
    font_size = font_ref$base_font_size,
    font_name = font_ref$name
  )

  wb
}

#' Apply Styles to a Sheet Title
#' @param wb An 'openxlsx2' wbWorkbook object.
#' @param tab_title Character. The tab in `wb` where the style should be set.
#' @param style_ref List. The style-reference object made with .style_paragraph().
#' @param font_ref List. The font-reference object made with .style_font().
#' @noRd
.style_sheet_title <- function(wb, tab_title, style_ref, font_ref) {
  # Sheet titles are BOLD and 16PT by default
  # .style_font() checks the config.yaml file for user preferences
  # which are then included here in font_ref
  wb$add_font(
    sheet = tab_title,
    dims = "A1",
    size = font_ref[["sheet_heading_size"]],
    bold = font_ref[["bold"]],
    name = font_ref[["name"]]
  )

  wb$add_cell_style(
    sheet = tab_title,
    dims = "A1",
    horizontal = style_ref[["lalign"]]
  )

  wb
}

#' Apply Styles to a Table
#' @param wb An 'openxlsx2' wbWorkbook object.
#' @param table_name Character. The table to which styles should be applied.
#' @param style_ref List. The style-reference object made with .style_paragraph().
#' @param font_ref List. The font-reference object made with .style_font().
#' @noRd

.style_table <- function(wb,
                         content,
                         table_name,
                         style_ref,
                         font_ref,
                         table_formats,
                         workbook_format) {

  content_row <- content[content[["table_name"]] == table_name, ]
  table <- content_row[, "table"][[1]]
  tab_title <- content_row[, "tab_title"][[1]]

  if (!is.null(workbook_format$cellwidth_default)) {
    cellwidth_default <- workbook_format$cellwidth_default
  } else {
    cellwidth_default <- 16
  }

  if (!is.null(workbook_format$cellwidth_wider)) {
    cellwidth_wider <- workbook_format$cellwidth_wider
  } else {
    cellwidth_wider <- 32
  }

  if (!is.null(workbook_format$nchar_break)) {
    nchar_break <- workbook_format$nchar_break
  } else {
    nchar_break <- 50
  }

  start_row <- .get_start_row_table(
    content,
    tab_title,
    .has_notes(content, tab_title),
    .has_blanks_message(content, tab_title),
    .has_custom_rows(content, tab_title),
    .has_source(content, tab_title)
  )

  table_height <- nrow(table)
  table_width <- ncol(table)

  #=============================================================================
  # style table headers, wrap text, left align columns by default
  #=============================================================================

  # Wrap text
  wb$add_cell_style(
    sheet = tab_title,
    dims = wb_dims(
      rows = seq(start_row, start_row + table_height),
      cols = seq(table_width)
    ),
    wrap_text = style_ref[["wrap_text"]]
  )

  # Left align text
  wb$add_cell_style(
    sheet = tab_title,
    dims = wb_dims(
      rows = seq(start_row, start_row + table_height),
      cols = seq(table_width)
    ),
    horizontal = style_ref[["lalign"]]
  )

  # Table headers are bold
  # .style_font() checks the config.yaml file for user preferences
  # which are then included here in font_ref
  wb$add_font(
    sheet = tab_title,
    dims = wb_dims(rows = start_row, cols = seq(table_width)),
    bold = font_ref[["bold"]],
    size = font_ref[["table_header_size"]],
    name = font_ref[["name"]]
  )

  #=============================================================================
  # Set column widths
  #=============================================================================

  # Find indices of columns that should be wider than default
  is_factor_column <- sapply(table, is.factor) # nchar (below) fails on factors
  table[is_factor_column] <- lapply(table[is_factor_column], as.character)
  wide_cells <- names(Filter(function(x) max(tidyr::replace_na(nchar(x), 0)) > nchar_break, table))
  wide_cells_index <- which(names(table) %in% wide_cells)
  wide_headers_index <- which(nchar(names(table)) > nchar_break)
  wide_cols_index <- unique(c(wide_cells_index, wide_headers_index))

  wb$set_col_widths(
    sheet = tab_title,
    cols = seq(table_width),
    widths = cellwidth_default
  )

  if (length(wide_cols_index) >= 1) {
    wb$set_col_widths(
      sheet = tab_title,
      cols = wide_cols_index,
      widths = cellwidth_wider
    )
  }

  #=============================================================================
  # right align numeric columns
  #=============================================================================

  # get the index of numeric columns
  numeric_cols_names <- table_formats$numeric_columns
  numeric_cols_index <- which(names(table) %in% numeric_cols_names)

  if (length(numeric_cols_index > 0)) {
    wb$add_cell_style(
      sheet = tab_title,
      dims = wb_dims(
        rows = seq(start_row, start_row + table_height),
        cols = numeric_cols_index
      ),
      horizontal = style_ref[["ralign"]]
    )
  }

  #=============================================================================
  # insert currency symbols and format numbers
  #=============================================================================

  if (!is.null(table_formats$number_formats)) {
    # apply numeric formatting to entire table
    table_formats$number_formats |>
      pwalk(\(cell_reference, cell_format) {
        wb$add_numfmt(
          sheet = tab_title,
          dims = cell_reference,
          numfmt = cell_format
        )
      })
  }

  wb
}

#' Apply Styles to the Cover Sheet
#' @param wb An 'openxlsx2' wbWorkbook object.
#' @param tab_title Character. The tab in `wb` where the style should be set.
#' @param style_ref List. The style-reference object made with .style_paragraph().
#' @param font_ref List. The font-reference object made with .style_font().
#' @noRd
.style_cover <- function(wb, content, style_ref, font_ref) {
  content_row <- content[content[["sheet_type"]] == "cover", ]
  tab_name <- content_row[, "tab_title"][[1]]
  table <- content_row[, "table"][[1]]

  # The cover column is SET-WIDTH

  wb$set_col_widths(
    sheet = tab_name,
    cols = 1,
    widths = 72
  )

  # The cover content can be provided as a list or data.frame
  cover_is_list <- inherits(table, "list")
  cover_is_df <- is.data.frame(table)

  if (cover_is_list) {
    table_vec <- unlist(c(rbind(names(table), table)))

    # The cover column is SET-WIDTH and WRAPPED

    table_height <- length(table_vec)

    wb$add_cell_style(
      sheet = tab_name,
      dims = wb_dims(rows = seq(table_height + 1), cols = 1),
      wrap_text = style_ref[["wrap_text"]]
    )

    # Also identify rows containing section headers
    subheader_rows <- which(table_vec %in% names(table)) + 1
  }

  if (cover_is_df) {
    # The cover column is WRAPPED

    table_height <- nrow(table)

    wb$add_cell_style(
      sheet = tab_name,
      dims = wb_dims(rows = seq(table_height * 2 + 1), cols = 1),
      wrap_text = style_ref[["wrap_text"]]
    )

    # Also identify rows containing section headers
    subheader_rows <- seq(2, table_height * 2, 2)
  }

  # Section header rows also have LARGER ROW HEIGHT, are BOLD and 14PT by default

  wb$set_row_heights(
    sheet = tab_name,
    rows = subheader_rows,
    heights = 34
  )

  wb$add_font(
    sheet = tab_name,
    dims = wb_dims(rows = subheader_rows, cols = 1),
    bold = font_ref[["bold"]],
    size = font_ref[["sheet_subheading_size"]],
    name = font_ref[["name"]]
  )

  wb
}

#' Apply Styles to the Contents Sheet
#' @param wb An 'openxlsx2' wbWorkbook object.
#' @param tab_title Character. The tab in `wb` where the style should be set.
#' @param style_ref List. The style-reference object made with .style_paragraph().
#' @noRd
.style_contents <- function(wb, content, style_ref) {
  tab_title <- content[content[["sheet_type"]] == "contents", "tab_title"][[1]]
  table <- content[content[["sheet_type"]] == "contents", "table"][[1]]

  table_height <- nrow(table)
  table_width <- ncol(table)

  start_row <- .get_start_row_table(
    content,
    tab_title,
    .has_notes(content, tab_title),
    .has_blanks_message(content, tab_title),
    .has_custom_rows(content, tab_title),
    .has_source(content, tab_title)
  )

  # Contents columns are SET-WIDTH, WRAPPED and LEFT ALIGNED

  wb$set_col_widths(
    sheet = tab_title,
    cols = 1,
    widths = 16
  )

  wb$set_col_widths(
    sheet = tab_title,
    cols = 2,
    widths = 56
  )

  wb$add_cell_style(
    sheet = tab_title,
    dims = wb_dims(rows = seq(start_row, table_height + start_row), cols = seq(table_width)),
    wrap_text = style_ref[["wrap_text"]],
    horizontal = style_ref[["lalign"]]
  )
}

#' Apply Styles to the Notes Sheet
#' @param wb An 'openxlsx2' wbWorkbook object.
#' @param tab_title Character. The tab in `wb` where the style should be set.
#' @param style_ref List. The style-reference object made with .style_paragraph().
#' @noRd
.style_notes <- function(wb, content, style_ref) {
  tab_title <- content[content[["sheet_type"]] == "notes", "tab_title"][[1]]
  table <- content[content[["sheet_type"]] == "notes", "table"][[1]]

  table_height <- nrow(table)
  table_width <- ncol(table)

  start_row <- .get_start_row_table(
    content,
    tab_title,
    .has_notes(content, tab_title),
    .has_blanks_message(content, tab_title),
    .has_custom_rows(content, tab_title),
    .has_source(content, tab_title)
  )

  # Notes columns are SET-WIDTH, WRAPPED and LEFT ALIGNED

  wb$set_col_widths(
    sheet = tab_title,
    cols = 1,
    widths = 16
  )

  wb$set_col_widths(
    sheet = tab_title,
    cols = 2,
    widths = 56
  )

  wb$add_cell_style(
    sheet = tab_title,
    dims = wb_dims(rows = seq(start_row, table_height + start_row), cols = seq(table_width)),
    wrap_text = style_ref[["wrap_text"]],
    horizontal = style_ref[["lalign"]]
  )

}

#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @import openxlsx2
#' @importFrom dplyr across
#' @importFrom dplyr mutate
#' @importFrom dplyr everything
#' @importFrom dplyr filter
#' @importFrom dplyr all_of
#' @importFrom tidyselect where
#' @importFrom rlang .data
#' @importFrom rlang '%||%'
#' @importFrom pillar tbl_sum
#' @importFrom purrr pwalk
#' @importFrom purrr pluck
#' @importFrom purrr pluck_exists
#' @importFrom purrr list_modify
#' @importFrom purrr map_depth
#' @importFrom purrr compact
#' @importFrom stringr str_extract
#' @importFrom stringr str_replace
#' @importFrom stringr str_replace_all
#' @importFrom tibble tibble
## usethis namespace: end
NULL

#' A Demo 'data.frame' Object
#'
#' A pre-created data.frame ready to be converted to an aftables-class object
#' with \code{\link{as_aftable}} and then an 'openxlsx2' wbWorkbook-class object
#' with \code{\link{generate_workbook}}.
#'
#' @format A data.frame with 6 rows and 7 columns:
#' \describe{
#'   \item{tab_title}{Character. Text to appear on each sheet's tab.}
#'   \item{sheet_type}{Character. The content type for each sheet: 'cover', 'contents', 'notes', or 'tables'.}
#'   \item{sheet_title}{Character. The title that will appear in cell A1 (top-left) of each sheet.}
#'   \item{blank_cells}{Character. An explanation for any blank cells in the table.}
#'   \item{custom_rows}{List-column of character vectors. Additional arbitrary pre-table information provided by the user.}
#'   \item{source}{Character. The origin of the data, if relevant.}
#'   \item{table}{List-column of data.frames (apart from the cover, which is a list) containing the statistical tables.}
#' }
"demo_df"

#' A Demo 'aftables' Object
#'
#' A pre-created 'aftables' object ready to be converted to an 'openxlsx2'
#' wbWorkbook-class object with \code{\link{generate_workbook}}.
#'
#' @format A data.frame with 6 rows and 7 columns:
#' \describe{
#'   \item{tab_title}{Character. Text to appear on each sheet's tab.}
#'   \item{sheet_type}{Character. The content type for each sheet: 'cover', 'contents', 'notes', or 'tables'.}
#'   \item{sheet_title}{Character. The title that will appear in cell A1 (top-left) of each sheet.}
#'   \item{blank_cells}{Character. An explanation for any blank cells in the table.}
#'   \item{custom_rows}{List-column of character vectors. Additional arbitrary pre-table information provided by the user.}
#'   \item{source}{Character. The origin of the data, if relevant.}
#'   \item{table}{List-column of data.frames (apart from the cover, which is a list) containing the statistical tables.}
#' }
"demo_aftable"

#' A Demo 'Workbook' Object
#'
#' A pre-created 'openxlsx2' wbWorkbook-class object generated from an
#' aftables-class object with \code{\link{generate_workbook}}.
#'
#' @format An 'openxlsx2' wbWorkbook-class object with 5 sheets.
"demo_workbook"

#' Clean Sheet Tab Titles
#' @param tab_titles Character vector. Names of tabs in the workbook.
#' @noRd
.check_tab_titles <- function(tab_titles) {
  tab_title_num_start <- grep("^\\d", unlist(tab_titles))

  if (length(tab_title_num_start) > 0) {
    stop(
      "Elements in tab_titles must not begin with a numeral (change ",
      .vector_to_sentence(tab_titles[tab_title_num_start]), ").",
      call. = FALSE
    )
  }
}

#' Clean Sheet Tab Titles
#' @param tab_titles Character vector. Names of tabs in the workbook.
#' @noRd
.clean_tab_titles <- function(tab_titles) {
  tab_titles_cleaned <- gsub("[^[:alnum:][:space:]_]", "", tab_titles)
  tab_titles_cleaned <- trimws(tab_titles_cleaned)
  tab_titles_cleaned <- gsub(" ", "_", tab_titles_cleaned)
  tab_titles_cleaned <- strtrim(tab_titles_cleaned, 31)

  before <- tab_titles[!tab_titles %in% tab_titles_cleaned]
  after <- tab_titles_cleaned[!tab_titles %in% tab_titles_cleaned]

  if (length(before > 0)) {
    warning(
      "These tab_titles have been cleaned automatically: ",
      paste(before, collapse = ", "),
      " (now ", paste(after, collapse = ", "), ").",
      call. = FALSE
    )
  }

  tab_titles_cleaned
}

#' Add a Missing Terminal Period to a String
#' @param text Character. A string.
#' @noRd
.append_period <- function(text) {
  last_char <- substr(text, nchar(text), nchar(text))

  for (i in seq_along(text)) {
    if (!is.na(last_char[i]) && last_char[i] != ".") {
      text[i] <- paste0(text[i], ".")
    }
  }

  text
}

#' Validate an 'aftable' Object
#' @param text x. An object with class 'aftable', likely created with
#'     \code{\link{create_aftable}}.
#' @noRd
.validate_aftable <- function(x) {
  names_req <- c(
    "tab_title",
    "sheet_type",
    "sheet_title",
    "blank_cells",
    "source",
    "custom_rows",
    "table"
  )
  names_count <- length(names_req)
  names_in <- names(x)

  # Must be of data.frame class
  if (!inherits(x, "data.frame")) {
    stop("Input must have class data.frame.", call. = FALSE)
  }

  # Must have particular dimensions (must have cover, contents table, at least)
  if (length(names_req) != length(x) || nrow(x) < 3) {
    stop(
      "Input must be a data.frame with ", names_count,
      " columns and at least 4 rows.",
      call. = FALSE
    )
  }

  # Column names must match expected format
  if (!all(names_req %in% names_in)) {
    stop(
      "Input data.frame does not have the required column names.",
      call. = FALSE
    )
  }

  # 'table' column class must be listcol
  if (!inherits(x[["table"]], "list")) {
    stop(
      "Column 'table' must be a listcol of data.frame objects.",
      call. = FALSE
    )
  }

  # 'custom_row' column class must be listcol
  if (!inherits(x[["custom_rows"]], "list")) {
    stop(
      "Column 'table' must be a listcol of character vectors",
      call. = FALSE
    )
  }

  # Class must be character for all columns except 'table' and 'custom_rows'

  char_cols <- x[, !names(x) %in% c("table", "custom_rows")]
  are_char_cols <- unlist(lapply(char_cols, is.character))

  if (!all(are_char_cols)) {
    stop(
      "All columns except 'table' and 'custom_rows' must be character class.",
      call. = FALSE
    )
  }

  # Content of 'table' listcol must be single data.frame objects (or cover list)
  if (!all(unlist(lapply(x[["table"]], is.list)))) {
    stop(
      "List-column 'table' must contain data.frame objects only. ",
      "The cover can also be of class 'list'.",
      call. = FALSE
    )
  }

  # Content of 'custom_row' listcol must be character
  if (!all(unlist(lapply(x[["custom_rows"]], is.character)))) {
    stop(
      "List-column 'custom_rows' must contain character vectors only. ",
      call. = FALSE
    )
  }

  # There must be cover and contents sheets
  if (sum(x[["sheet_type"]] %in% c("cover", "contents")) < 2) {
    stop(
      "The input data.frame must have sheet_type 'cover' and 'contents'.",
      call. = FALSE
    )
  }

  # There must be only one cover, contents and (optional) notes sheet
  if (
    sum(x[["sheet_type"]] == "cover") > 1 ||
      sum(x[["sheet_type"]] == "contents") > 1 ||
      sum(x[["sheet_type"]] == "notes") > 1
  ) {
    stop(
      "There can be only one 'cover', 'contents' and 'notes' in sheet_type.",
      call. = FALSE
    )
  }

  # There should be no empty rows, except in the blank_cells or source columns
  if (
    any(
      is.na(
        subset(x, select = c("tab_title", "sheet_type", "sheet_title", "table"))
      )
    )
  ) {
    stop(
      "Columns 'tab_title', 'sheet_type', 'sheet_title' and ",
      "'table' must not contain NA.",
      call. = FALSE
    )
  }

  # Each sheet_type must be only one of four types
  if (!all(x[["sheet_type"]] %in% c("cover", "contents", "notes", "tables"))) {
    stop(
      "'sheet_type' must be one of 'cover', 'contents', 'notes', 'tables'.",
      call. = FALSE
    )
  }

  # Each tab_title must be unique
  if (length(x[["tab_title"]]) != length(unique(tolower(x[["tab_title"]])))) {
    stop("Each 'tab_title' must be unique (case-insensitive).", call. = FALSE)
  }
}

#' Warn if an 'aftable' Has a Non-critical Problem
#' @param text x. An object with class 'aftable', likely created with
#'     \code{\link{create_aftable}}.
#' @noRd
.warn_aftable <- function(content) {
  # Warn about tab_title limitations

  tab_titles <- content$tab_title

  if (any(nchar(tab_titles) > 31)) {
    warning(
      "Each tab_title must be shorter than 31 characters.",
      call. = FALSE
    )
  }

  if (any(grepl("[^[:alnum:]_]", tab_titles))) {
    warning(
      "Each tab_title must contain only letters, numbers or underscores.",
      call. = FALSE
    )
  }


  # Warn about missing sources

  table_sources <- content[content$sheet_type == "tables", "source"]

  if (any(is.na(table_sources))) {
    warning(
      "One of your tables is missing a source statement.",
      call. = FALSE
    )
  }

  # Warn about possibly missing rows in the contents table

  contents_table <- content[content$sheet_type == "contents", "table"][[1]]

  if (nrow(content) != nrow(contents_table) + 2) {
    warning(
      "There are ", nrow(content) - 2, " tables but ",
      nrow(contents_table), " in the contents sheet.",
      call. = FALSE
    )
  }

  # Warn about notes (missing notes sheet, or notes in table)

  tables_sheets <- content[content$sheet_type == "tables", ]

  has_notes_sheet <- ifelse(
    nrow(content[content$sheet_type == "notes", ]) > 0,
    TRUE, FALSE
  )

  has_notes <-
    any(
      unlist(
        lapply(
          tables_sheets[, "tab_title"],
          function(x) .has_notes(content, x)
        )
      )
    )


  if (has_notes_sheet) {
    if (!has_notes) {
      warning(
        "You have a 'notes' sheet, but no notes in your tables.",
        call. = FALSE
      )
    }
  }

  if (has_notes) {
    if (!has_notes_sheet) {
      warning(
        "You have notes in your tables, but no 'notes' sheet.",
        call. = FALSE
      )
    }
  }

  # Warn about notes (note sheet present, but mismatches exist)

  if (has_notes_sheet) {
    notes_sheet <- content[content$sheet_type == "notes", ] # max of one notes sheet

    notes_sheet_vector <- notes_sheet[, "table"][[1]][[1]] # assumes first col has note values

    notes_sheet_values <- unlist( # e.g. get c(1, 2) from c("[note 1]", "[note 2]")
      regmatches(
        notes_sheet_vector,
        gregexpr("\\d", notes_sheet_vector, perl = TRUE) # extract numbers
      )
    )

    tables_sheet_notes <- sort(
      unique(
        unlist(
          lapply(
            tables_sheets$tab_title,
            function(x) .extract_note_values(content, x)
          )
        )
      )
    )

    not_in_tables <- setdiff(notes_sheet_values, tables_sheet_notes)
    not_in_notes <- setdiff(tables_sheet_notes, notes_sheet_values)

    if (has_notes_sheet && has_notes && length(not_in_notes) > 0) {
      warning(
        "Some notes are in the tables (",
        paste(not_in_notes, collapse = ", "),
        ") but are missing from the notes sheet.",
        call. = FALSE
      )
    }

    if (has_notes_sheet && has_notes && length(not_in_tables) > 0) {
      warning(
        "Some notes are in the notes sheet (",
        paste(not_in_tables, collapse = ", "),
        ") but are missing from the tables.",
        call. = FALSE
      )
    }
  }

  # Warn about blank cells in tables

  tables_list <- stats::setNames(
    tables_sheets[["table"]], tables_sheets[["tab_title"]]
  )

  tables_with_na_lgl <- unlist(
    lapply(
      tables_list,
      function(x) any(!stats::complete.cases(x))
    )
  )

  tables_with_na_names <- names(tables_with_na_lgl[tables_with_na_lgl])

  tables_with_blanks_reason <-
    tables_sheets[!is.na(tables_sheets$blank_cells), ][["tab_title"]]

  tables_with_na_no_reason <-
    setdiff(tables_with_na_names, tables_with_blanks_reason)

  tables_with_reason_no_na <-
    setdiff(tables_with_blanks_reason, tables_with_na_names)

  if (length(tables_with_na_no_reason) > 0) {
    warning(
      "You have blank cells in these tables but haven't provided a reason: ",
      paste(tables_with_na_no_reason, collapse = ", "), ".",
      call. = FALSE
    )
  }

  if (length(tables_with_reason_no_na) > 0) {
    warning(
      paste(
        "There's no blank cells in these tables,",
        "but you've provided a reason for blank cells: "
      ),
      paste(tables_with_reason_no_na, collapse = ", "), ".",
      call. = FALSE
    )
  }
}

test_that("aftable can be created by hand (with list for cover)", {
  # Uses demo_df, which has a list containing cover information in the
  # 'table' column.
  expect_warning(
    expect_warning(
      test_aftable <-
        create_aftable(
          tab_titles = demo_df$tab_title,
          sheet_types = demo_df$sheet_type,
          sheet_titles = demo_df$sheet_title,
          tables = demo_df$table
        ),
      "One of your tables is missing a source statement."
    ),
    "You have blank cells in these tables but haven't provided a reason: Table_1."
  )

  expect_s3_class(test_aftable, class = "aftable")
  expect_identical(class(test_aftable), c("aftable", "tbl", "data.frame"))

  expect_error(
    create_aftable(
      tab_titles   = demo_df$tab_title,
      sheet_types  = "x",
      sheet_titles = demo_df$sheet_title,
      tables       = demo_df$table
    ),
    "The input data.frame must have sheet_type 'cover' and 'contents'."
  )

})

test_that("aftable can be created by hand (with df for cover)", {

  expect_warning(
    expect_warning(
      test_aftable <-
        create_aftable(
          tab_titles = demo_df$tab_title,
          sheet_types = demo_df$sheet_type,
          sheet_titles = demo_df$sheet_title,
          tables = c(
            list(data.frame("Section 1" = "Text.", "Section 2" = "Text.")),
            demo_df$table[2:length(demo_df$table)]
          )
        ),
      "One of your tables is missing a source statement."
    ),
    "You have blank cells in these tables but haven't provided a reason: Table_1."
  )

  expect_s3_class(test_aftable, class = "aftable")
  expect_identical(class(test_aftable), c("aftable", "tbl", "data.frame"))

  expect_error(
    expect_warning(
      create_aftable(
        tab_titles   = demo_df$tab_title,
        sheet_types  = "x",
        sheet_titles = demo_df$sheet_title,
        tables       = demo_df$table
      )
    ),
    "The input data.frame must have sheet_type 'cover' and 'contents'."
  )

})

test_that("strings are not converted to factors", {

  expect_warning(
    expect_warning(
      test_aftable <-
        create_aftable(
          tab_titles = demo_df$tab_title,
          sheet_types = demo_df$sheet_type,
          sheet_titles = demo_df$sheet_title,
          tables = demo_df$table
        ),
      "One of your tables is missing a source statement."
    ),
    "You have blank cells in these tables but haven't provided a reason: Table_1."
  )

  classes <- unlist(lapply(test_aftable, class))

  expect_true(all(c("character", "list") %in% classes))
  expect_false(any("factor" %in% classes))
})

test_that("suitable objects can be coerced", {

  test_aftable <-
    as_aftable(demo_df)

  expect_s3_class(test_aftable, class = "aftable")
  expect_identical(class(test_aftable), c("aftable", "tbl", "data.frame"))

  expect_identical(is_aftable(test_aftable), TRUE)
  expect_identical(is_aftable("test_aftable"), FALSE)

  expect_true(is_aftable(test_aftable))
  expect_false(is_aftable(mtcars))
})

test_that("class validation works", {

  expect_length(
    as_aftable(demo_df),
    7
  )

  expect_error(as_aftable(1),
               "Input must be a data.frame with 7 columns and at least 4 rows.")
  expect_error(as_aftable("x"),
               "Input must be a data.frame with 7 columns and at least 4 rows.")
  expect_error(as_aftable(list()),
               "Input must be a data.frame with 7 columns and at least 4 rows.")
  expect_error(as_aftable(data.frame()),
               "Input must be a data.frame with 7 columns and at least 4 rows.")

  test_demo_df <- demo_df
  names(test_demo_df)[1] <- "foo"
  expect_error(as_aftable(test_demo_df),
               "Input data.frame does not have the required column names.")

  test_demo_df <- demo_df
  test_demo_df[["table"]] <- as.character(test_demo_df[["table"]])
  expect_error(as_aftable(test_demo_df),
               "Column 'table' must be a listcol of data.frame objects.")

  test_demo_df <- demo_df[, 1:4]
  expect_error(as_aftable(test_demo_df),
               "Input must be a data.frame with 7 columns and at least 4 rows.")

  test_demo_df <- demo_df[1, ]
  expect_error(as_aftable(test_demo_df),
               "Input must be a data.frame with 7 columns and at least 4 rows.")

  test_demo_df <- demo_df
  test_demo_df[test_demo_df$sheet_type %in% c("cover", "contents"), "sheet_type"] <- "foo"
  expect_error(as_aftable(test_demo_df),
               "The input data.frame must have sheet_type 'cover' and 'contents'.")

  test_demo_df <- demo_df
  test_demo_df[test_demo_df$tab_title == "Table_2", "sheet_type"] <- "foo"
  expect_error(as_aftable(test_demo_df),
               "'sheet_type' must be one of 'cover', 'contents', 'notes', 'tables'.")

  test_demo_df <- demo_df
  test_demo_df$sheet_type <- NA_character_
  expect_error(as_aftable(test_demo_df),
               "The input data.frame must have sheet_type 'cover' and 'contents'.")

  test_demo_df <- demo_df
  test_demo_df$custom_rows <- NA_character_
  expect_error(as_aftable(test_demo_df),
               "Column 'table' must be a listcol of character vectors")

  test_demo_df <- demo_df
  test_demo_df$custom_rows <- rep(list(1), nrow(test_demo_df))
  expect_error(as_aftable(test_demo_df),
               "List-column 'custom_rows' must contain character vectors only.")

  test_demo_df <- demo_df
  test_demo_df[test_demo_df$tab_title == "Table_2", "tab_title"] <-
    "Lorem_ipsum_dolor_sit_amet__consectetur_adipiscing"

  expect_warning(
    as_aftable(test_demo_df),
    "These tab_titles have been cleaned automatically:"
  )

  test_demo_df <- demo_df
  test_demo_df[test_demo_df$tab_title == "Table_2", "tab_title"] <- "!?"

  expect_warning(
    as_aftable(test_demo_df),
    "These tab_titles have been cleaned automatically:"
  )

  test_demo_df <- demo_df
  test_demo_df[test_demo_df$sheet_type == "notes", "table"][[1]] <-
    list(
      data.frame(
        "Note number" = "[note 1]",
        "Note text" = "US gallons.",
        check.names = FALSE
      )
    )

  expect_warning(
    as_aftable(test_demo_df),
    "Some notes are in the tables"
  )

})

test_that("summary method works", {
  test_demo_df <- suppressWarnings(as_aftable(demo_df))
  expect_output(summary(test_demo_df),
                "An aftable with 5 sheets:")
})

test_that("absence of note sheets doesn't prevent aftable formation", {
  test_demo_df <- demo_df[demo_df$sheet_type != "notes", ]

  expect_warning(
    expect_warning(
      test_aftable <- as_aftable(test_demo_df),
      "There are 2 tables but 3 in the contents sheet."
    ),
    "You have notes in your tables, but no 'notes' sheet."
  )

  expect_s3_class(test_aftable, "aftable")

  expect_warning(
    test_wb <- generate_workbook(test_aftable),
    "Some of the recommended workbook properties are missing."
  )

  expect_s3_class(test_wb, c("wbWorkbook", "R6"))
})

test_that("tab_titles with starting numeral will error", {
  demo_df[demo_df$tab_title == "Table_2", "tab_title"] <- "2_Table"

  expect_error(
    with(
      demo_df,
      create_aftable(
        tab_titles   = tab_title,
        sheet_types  = sheet_type,
        sheet_titles = sheet_title,
        tables       = table
      )
    ),
    "Elements in tab_titles must not begin with a numeral \\(change 2_Table\\)\\."
  )
})

test_that("tbl output looks as intended", {
  x <- create_aftable(
    tab_titles = LETTERS[1:3],
    sheet_types = c("cover", "contents", "tables"),
    sheet_titles = LETTERS[1:3],
    sources = c(NA_character_, NA_character_, "Source"),
    tables = list(
      data.frame(x = "x"),
      data.frame(tab = "x", title = "x"),
      mtcars
    )
  )

  expect_snapshot_output(as_aftable(x))
})

test_that("non-conforming tab_titles are cleaned", {
  expect_warning(
    create_aftable(
      tab_titles = c("cover", "contents", "Table 2"),
      sheet_types = c("cover", "contents", "tables"),
      sheet_titles = c("Cover", "Contents", "Table"),
      sources = "Source",
      tables = list(
        demo_df[["table"]][[1]],
        demo_df[["table"]][[2]][3, ],
        mtcars
      )
    ),
    "These tab_titles have been cleaned automatically: Table 2 \\(now Table_2\\)"
  )
})

test_that("tab_titles are unique", {
  demo_df[demo_df$tab_title == "Table_2", "tab_title"] <- "Table_1"

  expect_error(
    with(
      demo_df,
      create_aftable(
        tab_titles   = tab_title,
        sheet_types  = sheet_type,
        sheet_titles = sheet_title,
        tables       = table
      )
    )
  )
})

test_that("input other than data.frame is intercepted during validation", {
  expect_error(.validate_aftable("x"),
               "Input must have class data.frame.")

  test_demo_df <- demo_df
  test_demo_df[, "table"] <- "x"
  expect_error(
    as_aftable(test_demo_df),
    "Column 'table' must be a listcol of data.frame objects."
  )

  y <- subset(demo_df, select = -table)
  y[, "table"] <- list(rep(list("x"), nrow(y)))
  expect_error(
    as_aftable(y),
    "List-column 'table' must contain data.frame objects only."
  )
})

test_that("only one cover, contents, notes can be used", {
  cover_dupe <-
    rbind(demo_df, demo_df[demo_df$sheet_type == "cover", ])
  contents_dupe <-
    rbind(demo_df, demo_df[demo_df$sheet_type == "contents", ])
  notes_dupe <-
    rbind(demo_df, demo_df[demo_df$sheet_type == "notes", ])

  expect_error(as_aftable(cover_dupe))
  expect_error(as_aftable(contents_dupe))
  expect_error(as_aftable(notes_dupe))
})

test_that("NAs in certain columns cause failure", {
  tab_na <- demo_df
  tab_na$tab_title <- NA_character_
  type_na <- demo_df
  type_na$sheet_type <- NA_character_
  title_na <- demo_df
  title_na$sheet_title <- NA_character_
  table_na <- demo_df
  table_na$table <- NA_character_

  expect_error(as_aftable(tab_na))
  expect_error(as_aftable(type_na))
  expect_error(as_aftable(title_na))
  expect_error(as_aftable(table_na))
})

test_that("Note mismatch is caught", {
  test_demo_df <- demo_df[!demo_df$tab_title == "Table_1", ]
  test_demo_df[test_demo_df$sheet_type == "contents", "table"][[1]] <-
    list(data.frame(x = c("x", "y"), y = c("x", "y")))

  as_aftable(demo_df)

  test_demo_df <- demo_df[!demo_df$tab_title == "Table_2", ]
  test_demo_df[test_demo_df$sheet_type == "contents", "table"][[1]] <-
    list(data.frame(x = c("x", "y"), y = c("x", "y")))
  test_demo_df[test_demo_df$sheet_type == "notes", "table"][[1]] <- list(
    data.frame(
      `Note number` = paste0("[note ", 1:4, "]"),
      `Note text` = "x",
      check.names = FALSE
    )
  )

  expect_warning(
    as_aftable(test_demo_df),
    "Some notes are in the notes sheet"
  )
})


test_that("warning is raised if a source statement is missing", {
  demo_df[demo_df$tab_title == "Table_1", "source"] <- NA_character_

  expect_warning(
    as_aftable(demo_df),
    "One of your tables is missing a source statement."
  )
})

test_that("warning is raised if there's no blank cells but there is a reason", {
  demo_df[demo_df$tab_title == "Table_2", "blank_cells"] <- "x"

  expect_warning(
    as_aftable(demo_df),
    "There's no blank cells in these tables"
  )
})

test_that("period added to end of text if needed", {
  expect_identical(.append_period("Test"), "Test.")
  expect_identical(.append_period("Test."), "Test.")
  expect_identical(.append_period(NA), NA)

  expect_identical(
    .append_period(c("Test", "Test.", NA_character_)),
    c("Test.", "Test.", NA)
  )
})

test_that("tab titles are cleaned and warnings provided", {
  long_title <- "X12345678901234567890123456789012"

  expect_warning(
    .clean_tab_titles("Table 1"),
    "These tab_titles have been cleaned automatically: Table 1 \\(now Table_1\\)\\."
  )

  expect_warning(
    .clean_tab_titles(c("Table 1", "Table 2")),
    "These tab_titles have been cleaned automatically: Table 1, Table 2 \\(now Table_1, Table_2\\)\\."
  )

  expect_warning(
    .clean_tab_titles(c("Table_1", "Table 2")),
    "These tab_titles have been cleaned automatically: Table 2 \\(now Table_2\\)\\."
  )

  expect_warning(
    .clean_tab_titles("Table!@£#$%^&*(){}[]-=+;:'\"\\|<>,.~`/?1"),
    "These tab_titles have been cleaned automatically:.+now Table1\\)\\."
  )

  expect_warning(
    .clean_tab_titles(long_title),
    "These tab_titles have been cleaned automatically: X12345678901234567890123456789012 \\(now X123456789012345678901234567890\\)\\."
  )

  test_demo_df <- demo_df
  test_demo_df[1, "tab_title"] <- long_title

  expect_warning(
    as_aftable(test_demo_df),
    "These tab_titles have been cleaned automatically: X12345678901234567890123456789012 \\(now X123456789012345678901234567890\\)\\."
  )

  test_demo_df <- demo_df
  test_demo_df[1, "tab_title"] <- "Cover!"
  expect_warning(
    as_aftable(test_demo_df),
    "These tab_titles have been cleaned automatically: Cover! \\(now Cover\\)\\."
  )

  test_demo_df <- demo_df
  test_demo_df["tab_title"][5, ] <- long_title


  expect_warning(
    .warn_aftable(test_demo_df),
    "Each tab_title must be shorter than 31 characters."
  )

  test_demo_df <- demo_df
  test_demo_df["tab_title"][5, ] <- "Table-1!"

  expect_warning(
    .warn_aftable(test_demo_df),
    "Each tab_title must contain only letters, numbers or underscores."
  )
})

test_that("input column names are okay", {
  names(demo_df)[1] <- "x"
  expect_error(as_aftable(demo_df))
})

test_that("character class columns are caught if not character class", {
  demo_df[, "sheet_type"] <- seq_len(nrow(demo_df))
  expect_error(as_aftable(demo_df))
})

# Validate ----------------------------------------------------------------


.stop_bad_input <- function(wb, content, table_name = NULL) {
  if (!inherits(wb, "wbWorkbook")) {
    stop("'wb' must be an openxlsx2 wbWorkbook-class object.")
  }

  if (!is.null(table_name) &&
      !inherits(table_name, "character") &&
      length(table_name != 1)
  ) {
    stop("'table_name' must be a string of length 1")
  }
}


# Detect meta elements ----------------------------------------------------

.has_blanks_message <- function(content, tab_title) {
  blank_cells_message <- content[content$tab_title == tab_title, "blank_cells"][[1]]

  if (!is.na(blank_cells_message)) {
    TRUE
  } else {
    FALSE
  }
}

.has_source <- function(content, tab_title) {
  table_source <- content[content$tab_title == tab_title, "source"][[1]]

  if (!is.na(table_source)) {
    TRUE
  } else {
    FALSE
  }
}

.has_custom_rows <- function(content, tab_title) {
  custom_rows <- content[content$tab_title == tab_title, "custom_rows"][[1]]

  if (any(!is.na(custom_rows))) {
    TRUE
  } else {
    FALSE
  }
}

.has_notes <- function(content, tab_title) {
  table_names <- names(content[content$tab_title == tab_title, "table"][[1]])

  has_header_notes <- any(grepl("(?<=\\[).*(?=\\])", table_names, perl = TRUE))

  has_notes_column <- any(tolower(table_names) %in% "notes")

  any(has_header_notes, has_notes_column)
}

.extract_note_values <- function(content, tab_title) {
  has_notes <- .has_notes(content, tab_title)

  if (has_notes) { # if there are notes in this table

    # Isolate named table dataframe

    table <- content[content$tab_title == tab_title, "table"][[1]]

    # Vector with potential note values, e.g. '[1, 2]' to c(1, 2)

    table_names <- names(table)
    notes_column_index <- which(tolower(table_names) %in% "notes")
    notes_column_content <- table[, notes_column_index]

    possible_note_text <- c(table_names, notes_column_content)

    square_bracket_contents <- unlist(
      regmatches(
        possible_note_text,
        gregexpr("(?<=\\[).*(?=\\])", possible_note_text, perl = TRUE)
      )
    )

    sort(
      as.numeric(
        unique(
          unlist(
            lapply(
              square_bracket_contents,
              function(x) unlist(regmatches(x, gregexpr("\\d", x, perl = TRUE)))
            )
          )
        )
      )
    )
  }
}


# Table placement ---------------------------------------------------------


.get_start_row_blanks_message <- function(has_notes, start_row = 3) {
  if (has_notes) {
    start_row <- start_row + 1
  }

  start_row
}

.get_start_row_custom_rows <- function(has_notes,
                                       has_blanks_message,
                                       start_row = 3) {
  if (has_notes) {
    start_row <- start_row + 1
  }

  if (has_blanks_message) {
    start_row <- start_row + 1
  }

  start_row
}

.get_start_row_source <- function(content,
                                  tab_title,
                                  has_notes,
                                  has_blanks_message,
                                  has_custom_rows,
                                  start_row = 3) {
  if (has_notes) {
    start_row <- start_row + 1
  }

  if (has_blanks_message) {
    start_row <- start_row + 1
  }

  if (has_custom_rows) {
    custom_rows <- content[content$tab_title == tab_title, "custom_rows"][[1]]
    start_row <- start_row + length(custom_rows)
  }

  start_row
}

.get_start_row_table <- function(content,
                                 tab_title,
                                 has_notes,
                                 has_blanks_message,
                                 has_custom_rows,
                                 has_source,
                                 start_row = 3) {
  if (has_notes) {
    start_row <- start_row + 1
  }

  if (has_blanks_message) {
    start_row <- start_row + 1
  }

  if (has_custom_rows) {
    custom_rows <- content[content$tab_title == tab_title, "custom_rows"][[1]]
    start_row <- start_row + length(custom_rows)
  }

  if (has_source) {
    start_row <- start_row + 1
  }

  start_row
}


# Insert sheet elements ---------------------------------------------------


.insert_title <- function(wb, content, tab_title) {
  sheet_type <- content[content$tab_title == tab_title, "sheet_type"][[1]]
  sheet_title <- content[content$tab_title == tab_title, "sheet_title"][[1]]

  wb$add_data(
    sheet = tab_title,
    x = sheet_title,
    start_col = 1,
    start_row = 1,
    na.strings = ""
  )

  wb
}

.insert_table_count <- function(wb, content, tab_title) {
  table_count <- nrow(content[content$tab_title == tab_title, ])

  if (table_count < 10) {
    table_count <- switch(as.character(table_count),
      "1"  = "one",
      "2"  = "two",
      "3"  = "three",
      "4"  = "four",
      "5"  = "five",
      "6"  = "six",
      "7"  = "seven",
      "8"  = "eight",
      "9"  = "nine",
    )
  }

  text <- paste(
    "This worksheet contains", table_count,
    ifelse(table_count == "one", "table.", "tables.")
  )

  wb$add_data(
    sheet = tab_title,
    x = text,
    start_col = 1,
    start_row = 2, # table count will always be the second row,
    na.strings = ""
  )

  wb
}

.insert_notes_statement <- function(wb, content, tab_title) {
  has_notes <- .has_notes(content, tab_title)

  if (has_notes) {
    text <-
      "This table contains notes, which can be found in the Notes worksheet."

    wb$add_data(
      sheet = tab_title,
      x = text,
      start_col = 1,
      start_row = 3, # notes will always go in row 3 if they exist
      na.strings = ""
    )
  }

  wb
}

.insert_blanks_message <- function(wb, content, tab_title) {
  has_blanks_message <- .has_blanks_message(content, tab_title)

  if (has_blanks_message) {
    blanks_text <- content[content$tab_title == tab_title, "blank_cells"][[1]]
    has_notes <- .has_notes(content, tab_title)
    start_row <- .get_start_row_blanks_message(has_notes)

    wb$add_data(
      sheet = tab_title,
      x = blanks_text,
      start_col = 1,
      start_row = start_row,
      na.strings = ""
    )
  }

  wb
}

.insert_custom_rows <- function(wb, content, tab_title) {
  has_custom_rows <- .has_custom_rows(content, tab_title)

  if (has_custom_rows) {
    custom_rows_text <-
      content[content$tab_title == tab_title, "custom_rows"][[1]]

    custom_rows_text <- lapply(custom_rows_text, .make_hyperlink)

    has_notes <- .has_notes(content, tab_title)
    has_blanks <- .has_blanks_message(content, tab_title)
    start_row <- .get_start_row_custom_rows(has_notes, has_blanks)

    for (i in seq_along(custom_rows_text)) {
      has_hyperlink <- class(custom_rows_text[[i]]) == "hyperlink"

      if (has_hyperlink) {
        wb$add_formula(
          sheet = tab_title,
          x = create_hyperlink(
            text = names(custom_rows_text[[i]]),
            file = custom_rows_text[[i]]
          ),
          dims = wb_dims(cols = 1, rows = start_row + (i - 1))
        )
      }

      if (!has_hyperlink) {
        wb$add_data(
          sheet = tab_title,
          x = custom_rows_text[[i]],
          start_row = start_row + (i - 1),
          na.strings = ""
        )
      }
    }
  }

  wb
}

.insert_source <- function(wb, content, tab_title) {
  has_source <- .has_source(content, tab_title)

  if (has_source) {
    source_text <- content[content$tab_title == tab_title, "source"][[1]]
    source_text <- paste("Source:", source_text)
    source_text <- .make_hyperlink(source_text)

    start_row <- .get_start_row_source(
      content,
      tab_title,
      .has_notes(content, tab_title),
      .has_blanks_message(content, tab_title),
      .has_custom_rows(content, tab_title)
    )

    has_hyperlink <- class(source_text) == "hyperlink"
    if (has_hyperlink) {
      wb$add_formula(
        sheet = tab_title,
        x = create_hyperlink(
          text = names(source_text)[[1]],
          file = source_text
        ),
        dims = wb_dims(cols = 1, rows = start_row)
      )
    } else {
      wb$add_data(
        sheet = tab_title,
        x = source_text,
        start_col = 1,
        start_row = start_row,
        na.strings = ""
      )
    }
  }

  wb
}

.insert_table <- function(wb, content, table_name) {

  # convert tibbles to data frames before processing
  table <- as.data.frame(content[content$table_name == table_name, ][["table"]][[1]])
  sheet_type <- content[content$table_name == table_name, "sheet_type"][[1]]
  tab_title <- content[content$table_name == table_name, "tab_title"][[1]]

  start_row <- .get_start_row_table(
    content,
    tab_title,
    .has_notes(content, tab_title),
    .has_blanks_message(content, tab_title),
    .has_custom_rows(content, tab_title),
    .has_source(content, tab_title)
  )

  # Determine cell / column types

  table_datatypes <- .determine_table_datatypes(table)

  numeric_columns <- table_datatypes$numeric_columns

  currency_cells <- table_datatypes$currency_cells

  numeric_cells <- table_datatypes$numeric_cells

  note_cells <- table_datatypes$note_cells


  # Get table cell reference positions

  table_cell_references <-
    wb_dims(
      x = table,
      from_row = start_row,
      cols = names(table)
    )

  table_cell_references <- dims_to_rowcol(table_cell_references)

  table_cell_references <- t(outer(table_cell_references$col, table_cell_references$row, paste0))

  colnames(table_cell_references) <- names(table)


  if (length(numeric_columns) > 0) {

    #===========================================================================
    # Clean mixed columns by removing notes
    #===========================================================================

    note_values <- table[note_cells]
    note_cell_references <- table_cell_references[note_cells]

    notes_replacement <- data.frame(
      cell_reference = note_cell_references,
      cell_text = note_values
    )

    table[note_cells] <- ""


    #===========================================================================
    # Extract currency symbols from numeric columns for number formatting
    #===========================================================================
    currency_units <- .extract_currency_units(table, numeric_columns)

    #===========================================================================
    # Clean table removing currency symbols
    #===========================================================================
    table <- .replace_currency_units(table, numeric_columns)

    #===========================================================================
    # Convert numeric and currency columns to numeric
    #===========================================================================
    table <- .clean_numeric_data(table, numeric_columns)

    #===========================================================================
    # Determine decimal places from data for number formatting
    #===========================================================================
    decimal_places <- .determine_decimal_places(table, numeric_columns)

    #===========================================================================
    # Create number formats from currency units
    # and decimal places to pass to .style_table
    #===========================================================================

    number_cell_references <-
      table_cell_references[, numeric_columns, drop = FALSE]

    number_formats <- .determine_number_formats(
      currency_units,
      decimal_places,
      number_cell_references
    )

  } else {
    notes_replacement <-
      data.frame(
        cell_reference = character(0),
        cell_text = character(0)
      )

    number_formats <- NULL
  }

  #=============================================================================
  # insert data table into workbook
  #=============================================================================

  wb$add_data_table(
    sheet = tab_title,
    x = table,
    table_name = table_name,
    start_col = 1,
    start_row = start_row,
    table_style = "none",
    with_filter = FALSE,
    banded_rows = FALSE,
    na.strings = ""
  )

  #=============================================================================
  # insert notes into mixed columns
  #=============================================================================

  notes_replacement |>
    pwalk(\(cell_reference,
            cell_text) {
      wb$add_data(
        sheet = tab_title,
        x = cell_text,
        dims = cell_reference,
        col_names = FALSE,
        row_names = FALSE,
        apply_cell_style = FALSE
      )
    })

  #=============================================================================
  # create output to pass to .style_table
  #=============================================================================

  output <- list(
    numeric_columns = numeric_columns,
    number_formats = number_formats
  )

  output

}

# Special case to insert cover-page info, depending on whether it's provided as
# a df or list. All other tables in a workbook are provided as df only.
.insert_cover_table <- function(wb, content, table_name) {
  table <- content[content$table_name == "cover", ][["table"]][[1]]
  tab_title <- content[content$table_name == "cover", "tab_title"][[1]]

  if (inherits(table, "data.frame")) {
    table <- stats::setNames(
      as.list(table[["subsection_content"]]),
      table[["subsection_title"]]
    )
  }

  if (inherits(table, "list")) {
    table <- unlist(c(rbind(names(table), table)))
  }

  table_with_links <- lapply(table, .make_hyperlink)

  for (i in seq_along(table_with_links)) {
    has_hyperlink <- class(table_with_links[[i]]) == "hyperlink"

    if (has_hyperlink) {
      wb$add_formula(
        sheet = tab_title,
        x = create_hyperlink(
          text = names(table_with_links[[i]]),
          file = table_with_links[[i]]
        ),
        dims = wb_dims(cols = 1, rows = i + 1)
      )
    }

    if (!has_hyperlink) {
      wb$add_data(
        sheet = tab_title,
        x = table_with_links[[i]],
        start_row = i + 1,
        na.strings = ""
      )
    }
  }
}


# Handle hyperlinks -------------------------------------------------------


.detect_hyperlink <- function(string) {
  hyper_rx <- "\\[(([[:graph:]]|[[:space:]])+)\\]\\([[:graph:]]+\\)"
  grepl(hyper_rx, string)
}

.detect_multi_hyperlink <- function(string) {
  md_rx <- "\\[(([[:graph:]]|[[:space:]])+?)\\]\\([[:graph:]]+?\\)"
  md_match <- gregexpr(md_rx, string, perl = TRUE)
  md_extract <- regmatches(string, md_match)[[1]]
  has_multi_hyperlink <- length(md_extract) > 1

  if (has_multi_hyperlink) {
    warning(
      "String has more than one hyperlink, only first will be extracted.",
      call. = FALSE
    )
  }

  invisible(has_multi_hyperlink)
}

.check_scheme <- function(string) {
  scheme_rx <- paste("((http(s?)|ftp)://?)", "(mailto:?)", sep = "|")
  grepl(scheme_rx, string)
}

.extract_hyperlink <- function(string, keep_full_string = TRUE) {
  md_rx <- "\\[(([[:graph:]]|[[:space:]])+?)\\]\\([[:graph:]]+?\\)"
  md_match <- regexpr(md_rx, string, perl = TRUE)
  md_extract <- regmatches(string, md_match)[[1]]

  url_rx <- "(?<=\\]\\()([[:graph:]]|[[:space:]])+(?=\\))"
  url_match <- regexpr(url_rx, md_extract, perl = TRUE)
  url_extract <- regmatches(md_extract, url_match)[[1]]

  string_rx <- "(?<=\\[)([[:graph:]]|[[:space:]])+(?=\\])"
  string_match <- regexpr(string_rx, md_extract, perl = TRUE)
  string_extract <- regmatches(md_extract, string_match)[[1]]

  if (keep_full_string) {
    string_extract <- gsub(md_rx, string_extract, string)
  }

  named_hyperlink <- stats::setNames(url_extract, string_extract)
  class(named_hyperlink) <- "hyperlink"
  named_hyperlink
}

.make_hyperlink <- function(string) {
  has_hyperlink <- .detect_hyperlink(string)

  if (has_hyperlink) {
    .detect_multi_hyperlink(string)
    scheme_is_ok <- .check_scheme(string)

    if (scheme_is_ok) {
      string <- .extract_hyperlink(string)
    }
  }

  string
}


# Add sheets to workbook --------------------------------------------------


.add_tabs <- function(wb, content) {
  .stop_bad_input(wb, content)

  for (i in unique(content$tab_title)) {
    wb$add_worksheet(i)
  }

  wb
}

.add_cover <- function(wb, content, font_ref) {
  .stop_bad_input(wb, content)

  tab_title <- content[content$sheet_type == "cover", "tab_title"][[1]]
  table_name <- content[content$sheet_type == "cover", "table_name"][[1]]

  .insert_title(wb, content, tab_title)
  .insert_cover_table(wb, content, table_name) # rather than .insert_table

  styles <- .style_paragraph()
  .style_sheet_title(wb, tab_title, styles, font_ref)
  .style_cover(wb, content, styles, font_ref)
  # TODO: needs special handling if list provided
  wb
}


.add_contents <- function(wb, content, font_ref, workbook_format) {
  .stop_bad_input(wb, content)

  tab_title <- content[content$sheet_type == "contents", "tab_title"][[1]]
  table_name <- content[content$sheet_type == "contents", "table_name"][[1]]

  .insert_title(wb, content, tab_title)
  .insert_table_count(wb, content, tab_title)
  .insert_custom_rows(wb, content, tab_title)
  table_format <- .insert_table(wb, content, table_name)

  styles <- .style_paragraph()
  .style_sheet_title(wb, tab_title, styles, font_ref)
  .style_table(wb, content, table_name, styles, font_ref, table_format, workbook_format)
  .style_contents(wb, content, styles)

  wb
}


.add_notes <- function(wb, content, font_ref, workbook_format) {
  .stop_bad_input(wb, content)

  tab_title <- content[content$sheet_type == "notes", "tab_title"][[1]]
  table_name <- content[content$sheet_type == "notes", "table_name"][[1]]

  .insert_title(wb, content, tab_title)
  .insert_table_count(wb, content, tab_title)
  .insert_custom_rows(wb, content, tab_title)
  table_format <- .insert_table(wb, content, table_name)

  styles <- .style_paragraph()
  .style_sheet_title(wb, tab_title, styles, font_ref)
  .style_table(wb, content, table_name, styles, font_ref, table_format, workbook_format)
  .style_notes(wb, content, styles)

  wb
}

.add_tables <- function(wb, content, table_name, font_ref, workbook_format) {
  .stop_bad_input(wb, content, table_name)

  tab_title <- content[content$table_name == table_name, "tab_title"][[1]]

  .insert_title(wb, content, tab_title)
  .insert_table_count(wb, content, tab_title)
  .insert_source(wb, content, tab_title)
  .insert_notes_statement(wb, content, tab_title)
  .insert_blanks_message(wb, content, tab_title)
  .insert_custom_rows(wb, content, tab_title)
  table_format <- .insert_table(wb, content, table_name)

  styles <- .style_paragraph()
  .style_sheet_title(wb, tab_title, styles, font_ref)
  .style_table(wb, content, table_name, styles, font_ref, table_format, workbook_format)

  wb
}

.determine_empty_cells <- function(table) {

  empty_cells <- table |>
    mutate(across(everything(), \(x) is.na(x) | x == ""))

  empty_cells
}

.determine_currency_cells <- function(table) {
  currency_cells <-
    table |>
    mutate(
      across(
        everything(),
        \(x) {
          grepl(x,
                pattern = detect_currency_regex,
                perl = TRUE)
        }
      )
    )

  currency_cells
}

.determine_numeric_cells <- function(table) {
  numeric_cells <-
    table |>
    mutate(
      across(
        everything(),
        \(x) {
          grepl(x, pattern = numeric_regex, perl = TRUE)
        }
      )
    )

  numeric_cells
}

.determine_note_cells <- function(table) {
  note_cells <-
    table |>
    mutate(
      across(
        everything(),
        \(x) {
          grepl(x, pattern = notes_regex, perl = TRUE)
        }
      )
    )

  note_cells
}

.determine_table_datatypes <- function(table) {

  # switch off scientific notation
  old <- options(scipen = 999)
  on.exit(options(old), add = TRUE)

  # number of cells in each column of each type
  currency_cells <- .determine_currency_cells(table)

  numeric_cells <- .determine_numeric_cells(table)

  empty_cells <- .determine_empty_cells(table)

  note_cells <-  .determine_note_cells(table)

  # total number of cells in each column of each type
  currency_cells_count <- sapply(currency_cells, function(x) sum(x))
  numeric_cells_count <- sapply(numeric_cells, function(x) sum(x))
  note_cells_count <- sapply(note_cells, function(x) sum(x))
  empty_cells_count <- sapply(empty_cells, function(x) sum(x))

  # columns which should be treated as numeric can contain either currency or
  # numeric data and character data in note format between square brackets []
  # numeric columns can have any number of note cells but should not be
  # entirely note or empty cells
  # columns which are entirely note or empty cells are note columns and should
  # be treated as characters
  # a column will only be counted if all the cells can be classified as one
  # of the data types, to avoid cases of unusual data being treated as numeric
  numeric_columns <-
    (
      (currency_cells_count >= 1 | numeric_cells_count >= 1) &
      (currency_cells_count +
       numeric_cells_count +
       note_cells_count +
       empty_cells_count) == nrow(table)
    )

  # valid currency cells are only those in columns which could be numeric
  # if all currency symbols and notes were removed
  currency_cells[!numeric_columns] <- FALSE

  # valid numeric cells are only those in columns which could be numeric
  # if all notes were removed
  numeric_cells[!numeric_columns] <- FALSE

  # valid note cells are only those in mixed columns which could be numeric
  # if all notes were removed
  note_cells[!numeric_columns] <- FALSE


  output <- list(
    numeric_columns = names(numeric_columns[numeric_columns]),
    currency_cells = as.matrix(currency_cells),
    numeric_cells = as.matrix(numeric_cells),
    note_cells = as.matrix(note_cells)
  )

  output
}

.extract_numeric_values <- function(values) {
  numeric_values <-
    values[
      sapply(
        values,
        grepl,
        pattern = numeric_regex,
        perl = TRUE
      )
    ]

  numeric_values <- str_replace_all(numeric_values, "[\\s,]", "")

  numeric_values <- as.numeric(numeric_values)

  numeric_values
}

.extract_currency_units <- function(table,
                                    numeric_columns) {

  output <- table |>
    dplyr::select(all_of(numeric_columns)) |>
    mutate(
      across(
        everything(),
        \(x) {
          tidyr::replace_na(str_extract(x, extract_currency_symbol_regex), "")
        }
      )
    )

  output

}

.replace_currency_units <- function(table,
                                    numeric_columns) {

  output <- table |>
    mutate(
      across(
        where(\(x) !is.numeric(x)) & all_of(numeric_columns),
        \(x) {
          str_replace(x, extract_currency_symbol_regex, "")
        }
      )
    )

  output
}

.clean_numeric_data <- function(table,
                                numeric_columns) {

  output <- table |>
    mutate(
      across(
        where(\(x) !is.numeric(x)) & all_of(numeric_columns),
        \(x) as.numeric(str_replace_all(x, "[\\s,]", ""))
      )
    )

  output
}

.set_workbook_properties <- function(wb, content) {

  if (!is.null(content$keywords)) {
    content$keywords <- paste0(content$keywords, collapse = ", ")
  }

  wb$set_properties(
    creator = content$author,
    title = content$title,
    subject = content$subject,
    category = content$category,
    keywords = content$keywords,
    comments = content$comments
  )

  wb
}

.determine_decimal_places <- function(table, numeric_columns) {

  # switch off scientific notation
  old <- options(scipen = 999)
  on.exit(options(old), add = TRUE)

  output <- table |>
    dplyr::select(all_of(numeric_columns)) |>
    mutate(
      across(everything(), \(x) nchar(abs(x)) - 1 - nchar(floor(abs(x)))),
      across(everything(), \(x) ifelse(x < 0, 0, x))
    ) |>
    dplyr::summarise(across(everything(), \(x) max(x, na.rm = TRUE)))

  output

}

.determine_number_formats <- function(currency_units,
                                      decimal_places,
                                      number_cell_references) {

  output <- list(
    cell_reference = as.vector(number_cell_references),
    # combine currency units (by table cell)
    # with decimal places (by numeric table column)
    cell_format = purrr::map2(
      currency_units,
      purrr::map(colnames(number_cell_references),
                 \(x) purrr::pluck(decimal_places, x)),
      \(currency_unit, decimal_length) {
        paste0(
          currency_unit,
          # default formatting with thousand separator
          "#,##0",
          # add decimal point if required
          if (decimal_length > 0) ".",
          # add number of digits after decimal point from decimal_length
          paste0(rep("0", decimal_length), collapse = "")
        )
      }
    ) |>
      unlist(use.names = FALSE)
  )

  output
}

test_that("vector elements are converted to sentence form", {
  expect_equal(.vector_to_sentence(LETTERS[1]), "A")
  expect_equal(.vector_to_sentence(LETTERS[1:2]), "A and B")
  expect_equal(.vector_to_sentence(LETTERS[1:3]), "A, B and C")
})


test_that(".determine functions correctly identify cells", {

  determine_functions_df <- data.frame(

    # Columns will all empty cells should remain the same
    empty = c(rep("", 8), NA),

    # Numeric columns should stay numeric
    numeric1 = c(1, 5, 26.25, 123, -15677, 23.45, 67, 45, NA),
    numeric2 = c(1:7, -8001L, NA),

    # Character columns containing only numbers should be converted to numeric
    number_character = c("1", "-1.345", "2322.456", "12.44", NA, "- 23.546",
                         " - 23", " 232.3 ", "0.12 "),

    # Character columns with mixed numbers and text in same cell should not be
    # converted to numeric
    mixed_text_numbers = c("12", "-15.6", "-13.4 [u]", " 1800.1 [note2]", NA,
                           "[c]", "13.7[u][bc] ", "78.6 [note 1][Note 4]",
                           "12.7 [note1, note2]"),

    # Character columns with numbers and note cells should be converted to
    # numeric
    number_notes = c("1", " -3.3", "-2834.459 ", "-23,456", "[c]",
                     "[-0.3cs] [g f]", "[ as][f]", "[dvb12,v.]", "7.8"),

    # Columns containing money should be converted to numeric
    # Columns with money and notes/numbers should be converted to numeric
    currency_single = c(NA, "-£12.3", "£13.556", "£0.6", "£-13", "£15001",
                        "£19,000.12", " £ 12.3", "£- 12 "), # currency, all £
    currency_multiple = c("£12.3", NA, "-£ 13.559", "$0.6", "£13", " £15,001",
                          " €19,000.12", " £12.3",
                          "12 "), # Currency, mixed symbols
    currency_notes = c("£12.3", NA, "£ 13.55", "-$0.6", " [-£0.2] [cvb] ",
                       " £15,001", " €19,000.12", " £12.3",
                       "12 $"), # Currency and notes

    # Columns containing only notes should stay as character
    notes = c("[a]", "[abv] [efg]", " [note 1]", "[c]  ", "[abc][efg]",
              "[ab cd]", "[a][b][c]", "[a$]", "[12]")

  )

  expected_notes_cells <- data.frame(
    empty = rep(FALSE, 9),
    numeric1 = rep(FALSE, 9),
    numeric2 = rep(FALSE, 9),
    number_character = rep(FALSE, 9),
    mixed_text_numbers = c(rep(FALSE, 5), TRUE, rep(FALSE, 3)),
    number_notes = c(rep(FALSE, 4), rep(TRUE, 4), FALSE),
    currency_single = rep(FALSE, 9),
    currency_multiple = rep(FALSE, 9),
    currency_notes = c(rep(FALSE, 4), TRUE, rep(FALSE, 4)),
    notes = rep(TRUE, 9)
  )

  expected_numeric_cells <- data.frame(
    empty = rep(FALSE, 9),
    numeric1 = c(rep(TRUE, 8), FALSE),
    numeric2 = c(rep(TRUE, 8), FALSE),
    number_character = c(rep(TRUE, 4), FALSE, rep(TRUE, 4)),
    mixed_text_numbers = c(rep(TRUE, 2), rep(FALSE, 7)),
    number_notes = c(rep(TRUE, 4), rep(FALSE, 4), TRUE),
    currency_single = rep(FALSE, 9),
    currency_multiple = c(rep(FALSE, 8), TRUE),
    currency_notes = rep(FALSE, 9),
    notes = rep(FALSE, 9)
  )

  expected_currency <- data.frame(
    empty = rep(FALSE, 9),
    numeric1 = rep(FALSE, 9),
    numeric2 = rep(FALSE, 9),
    number_character = rep(FALSE, 9),
    mixed_text_numbers = rep(FALSE, 9),
    number_notes =  rep(FALSE, 9),
    currency_single = c(FALSE, rep(TRUE, 8)),
    currency_multiple = c(TRUE, FALSE, rep(TRUE, 6), FALSE),
    currency_notes = c(TRUE, FALSE, rep(TRUE, 2), FALSE, rep(TRUE, 4)),
    notes = rep(FALSE, 9)
  )

  expected_empty <- data.frame(
    empty = rep(TRUE, 9),
    numeric1 = c(rep(FALSE, 8), TRUE),
    numeric2 = c(rep(FALSE, 8), TRUE),
    number_character = c(rep(FALSE, 4), TRUE, rep(FALSE, 4)),
    mixed_text_numbers = c(rep(FALSE, 4), TRUE, rep(FALSE, 4)),
    number_notes =  rep(FALSE, 9),
    currency_single = c(TRUE, rep(FALSE, 8)),
    currency_multiple = c(FALSE, TRUE, rep(FALSE, 7)),
    currency_notes = c(FALSE, TRUE, rep(FALSE, 7)),
    notes = rep(FALSE, 9)
  )

  notes_cells <- .determine_note_cells(determine_functions_df)
  expect_equal(notes_cells, expected_notes_cells)

  numeric_cells <- .determine_numeric_cells(determine_functions_df)
  expect_equal(numeric_cells, expected_numeric_cells)

  currency_cells <- .determine_currency_cells(determine_functions_df)
  expect_equal(currency_cells, expected_currency)

  empty_cells <- .determine_empty_cells(determine_functions_df)
  expect_equal(empty_cells, expected_empty)

  # numeric_columns identifies all columns which need to be output as numeric,
  # including columns containing currency and mix of numbers/currency and notes
  numeric_columns <-
    .determine_table_datatypes(determine_functions_df)$numeric_columns
  expect_equal(
    numeric_columns,
    c("numeric1", "numeric2", "number_character", "number_notes",
      "currency_single", "currency_multiple", "currency_notes")
  )

})


test_that("table cleaning functions work as intended", {

  # .replace_currency_units ----------------------------------------------------

  currency_df <- data.frame(
    col1 = c("£12.30", " 13$ ", "   €123,123.1234"),
    col2 = c(" €19,000.12", "£12.30", "£12 ")
  )

  expected_currency_df <- data.frame(
    col1 = c("12.30", " 13 ", "   123,123.1234"),
    col2 = c(" €19,000.12", "£12.30", "£12 ")
  )

  currency_df <- .replace_currency_units(currency_df, "col1")

  expect_equal(currency_df, expected_currency_df)


  # .clean_numeric_data --------------------------------------------------------

  df <- data.frame(
    col1 = c("12.30", NA, " 13 ", "   123,123.1234"),
    col2 = c(" 19,000.12", "12.30", "12 ", NA),
    col3 = c(123, 235, NA, 12.4),
    col4 = c("  123", "12,001 ", NA, " 12.1")
  )

  cleaned_df_expected <- data.frame(
    col1 = c(12.30, NA, 13, 123123.1234),
    col2 = c(19000.12, 12.30, 12, NA),
    col3 = c(123, 235, NA, 12.4),
    col4 = c("  123", "12,001 ", NA, " 12.1")
  )


  # Clean 3 of the 4 columns
  # Col 1 and 2 should be converted to numeric. Col 3 and 4 should remain the
  # same.
  cleaned_df <- .clean_numeric_data(df, c("col1", "col2", "col3")
  )

  expect_equal(cleaned_df, cleaned_df_expected)


})

---
title: "Introduction to {aftables}"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{aftables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)
```

## Purpose

This vignette demonstrates how to use {aftables} to generate a spreadsheet output that automatically follows [best practice](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/).

## Installation

Install the latest release version of aftables directly from CRAN:

``` {r install from CRAN, eval=FALSE}
install.packages("aftables")
```

## Workflow

Having installed the {aftables} package, there are three steps to generating a compliant spreadsheet:

1. Use `create_aftable()` to create a special dataframe (with class 'aftable') that contains all the content that will go in your spreadsheet
1. Pass the output to `generate_workbook()` to convert the aftable to {openxlsx2}'s 'wbWorkbook' class, which adds spreadsheet structure and styles
1. Pass the output to `openxlsx2::wb_save()` to write out to an xlsx file (or `openxlsx2::wb_open()` to open a temporary copy)

You can use the package's [RStudio Addin](https://rstudio.github.io/rstudioaddins/), which is installed with {aftables}, to insert a pre-filled demo skeleton of this workflow (RStudio users only).

### 1. Create an aftable

Each argument to `create_aftable()` provides the information needed to construct each sheet in the spreadsheet.

| Argument | Required | Type | Accepted values | Explanation |
| :--- | :- | :--- | :--- | :------ |
| `tab_titles` | Yes | Character vector | | The name that will appear on each sheet's tab in the output spreadsheet |
| `sheet_types` | Yes | Character vector | 'cover', 'contents', 'notes', 'tables' | The kind of information that the sheet holds, which is needed so that the correct structure and formatting can be applied later |
| `sheet_titles` | Yes | Character vector | | The main heading of each sheet, which will appear in cell A1 |
| `blank_cells` | No | Character vector | | A sentence that explains the reason for any blank cells in the sheet (if applicable) |
| `custom_rows` | No | List of character vectors | | Arbitrary rows of text that the user wants to insert above a table, one list-item per sheet (contents, notes and tables sheets), one vector element per row |
| `sources` | No | Character vector | | A sentence provides the source of the data found in each table (if applicable, likely only needed for sheets with `sheet_types` of 'table') |
| `tables` | Yes | List of dataframes (although the cover sheet content can be provided as a list object) | | The main content for each sheet, expressed as flat ([probably tidy](https://www.jstatsoft.org/article/view/v059i10)) dataframes of rows and columns (though the cover can be a list) |

You can read more about these arguments and their requirements in the function's help pages, which you can access by running `?create_aftable` in the R console. See also the [terminology vignette](https://best-practice-and-impact.github.io/aftables/articles/terminology), `vignette("terminology", "aftables")`, for these terms and more.

#### Pre-prepare tables

Rather than pass a big list of dataframes directly to the `tables` argument of `create_aftable()`, it's preferable to prepare them first into their own named objects.

Below are some demo tables that we will later pass to `create_aftable()`. I've used `tibble::tribble()` for human-readable row-by-row dataframe construction, but you can just use `data.frame()` if you want.

##### Meta-sheets

The cover can accept either a list or a data.frame (the latter was the only acceptable input prior to version 0.2.0). We recommend a list so that you can have multiple rows per section on the cover. This also means you can dedicate certain rows to be hyperlinks to web URLs or mailto links that will open an email client. Here's a demo list for the contents page (required):

```{r tables-cover-list}
cover_list <- list(
  "Section 1" = c("First row of Section 1.", "Second row of Section 1."),
  "Section 2" = "The only row of Section 2.",
  "Section 3" = c(
    "[Website](https://best-practice-and-impact.github.io/aftables/)",
    "[Email address](mailto:fake.address@aftables.com)"
  )
)
```

Note: a list is the preferred method of input for the cover. Previously, a data.frame was the only way to supply the data for the cover sheet in version 0.1 of the package.

Here's a demo table for the contents page (required):

```{r tables-contents}
contents_df <- data.frame(
  "Sheet name" = c("Notes", "Table_1", "Table_2"),
  "Sheet title" = c(
    "Notes used in this workbook",
    "First Example Sheet",
    "Second Example Sheet"
  ),
  check.names = FALSE
)
```

And here's a demo table for the notes page (not required if there's no notes in your tables), which has a column for the note number in the form '[note x]' and a column for the note itself:

```{r tables-notes}
notes_df <- data.frame(
  "Note number" = paste0("[note ", 1:3, "]"),
  "Note text" = c("First note.", "Second note.", "Third note."),
  check.names = FALSE
)
```

<details><summary>Click to preview these objects</summary>
```{r tables-meta-expand}
cover_list
contents_df
notes_df
```
</details>

##### Statistical tables

The code below generates a demo data.frame that we're going to pretend is the statistical data that we want to publish. It has columns with different sorts of data that we might want to publish. It also has suppressed values (e.g. '[c]' meaning 'confidential' data) and includes notes (in the form '[note x]').

```{r stats-df-1}
table_1_df <- data.frame(
  Category = LETTERS[1:10],
  "Numeric [note 1]" = 1:10,
  "Numeric suppressed" = c(1:4, "[c]", 6:9, "[x]"),
  "Numeric thousands" = abs(round(rnorm(10), 4) * 1e5),
  "Numeric decimal" = abs(round(rnorm(10), 5)),
  "Long name that means that the column width needs to be widened" = 1:10,
  Notes = c("[note 1]", rep(NA_character_, 4), "[note 2]", rep(NA_character_, 4)),
  check.names = FALSE
)
```

We'll create a second, simpler table as well, which will go on a separate sheet:

```{r stats-df-2}
table_2_df <- data.frame(Category = LETTERS[1:10], Numeric = 1:10)
```

<details><summary>Click to preview these tables of statistical data</summary>
```{r table-stat-expand}
table_1_df
table_2_df
```
</details>

See [the best practice guidance](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/) for more information on how to present data in these tables.

#### Create aftable

Now we can construct an aftable by passing the required sheet elements as character vectors with `c()`—or a `list()` in the case of the `tables` and `custom_rows` arguments—to the `create_aftable()` function. 

Note that:

* the element index of the object supplied to each argument is the sheet that it will be applied to (e.g. the tab title of the first sheet will be 'Cover', the sheet type of the second sheet will be 'contents' and the fourth sheet will contain the `table_1_df` table)
* you must use `NA_character_` wherever an element isn't required (e.g. there is no information about blank cells nor sources for the first three sheets)
* you can insert a template of this demo using the package's RStudio Addin

```{r new-aftable}
my_aftable <- aftables::create_aftable(
  tab_titles = c("Cover", "Contents", "Notes", "Table 1", "Table_2"),
  sheet_types = c("cover", "contents", "notes", "tables", "tables"),
  sheet_titles = c(
    "The 'aftables' Demo Workbook",
    "Table of contents",
    "Notes",
    "Table 1: First Example Sheet",
    "Table 2: Second Example Sheet"
  ),
  blank_cells = c(
    rep(NA_character_, 3),
    "Blank cells indicate that there's no note in that row.",
    NA_character_
  ),
  custom_rows = list(
    NA_character_,
    NA_character_,
    "A custom row.",
    c(
      "First custom row [with a hyperlink.](https://best-practice-and-impact.github.io/aftables/)",
      "Second custom row."
    ),
    "A custom row."
  ),
  sources = c(
    rep(NA_character_, 3),
    "[The Source Material., 2024](https://best-practice-and-impact.github.io/aftables/)",
    "The Source Material, 2024."
  ),
  tables = list(cover_list, contents_df, notes_df, table_1_df, table_2_df)
)
```

The function will return errors or warnings if anything is missing or seems odd. For example, we were warned that a value we supplied to `tab_title` had to be cleaned from 'Table 1' to 'Table_1', since blank spaces are not allowed in tab names. Note that there will be an error if there are any tab titles that start with a numeral.

Here's a preview of the object that was created:

```{r aftable-preview}
my_aftable
```

You can immediately tell that this is an aftable because it's the first word that's printed with the output.

So our aftable is basically just a table with one row per sheet and one column per sheet element. In fact, it has class 'data.frame'/'tbl' along with 'aftable'. For convenience, you can also check for the aftable class with `is_aftable()`.

Note that `create_aftable()` is the preferred method for generating aftable-class objects, but it's also possible to convert a correctly-formatted, pre-built data.frame or tibble directly to an aftable with `as_aftable()`.

### 2. Convert to a workbook

We can use `generate_workbook()` to convert our aftable to an [{openxlsx2}](https://janmarvin.github.io/openxlsx2/) wbWorkbook-class object.

This type of object adds information from each row of our aftable into separate sheets and applies other relevant structure, mark-up and styles for compliance with [the best practice guidance](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/).

```{r create-af-wb}
my_wb <- aftables::generate_workbook(my_aftable)
```

<details><summary>Click for a preview of the Workbook object</summary>

The print method for a Workbook-class object is fairly limited, but you can see an overview of our named sheets and some of the custom styling.

```{r wb-preview}
my_wb
```
</details>

### 3. Write to file

Finally, you can use [the `wb_save()` function from {openxlsx2}](https://janmarvin.github.io/openxlsx2/reference/wb_save.html) to write your workbook object to an xlsx file (set the path argument to a location suitable for your work).

```{r saveworkbook, eval=FALSE}
openxlsx2::wb_save(wb = my_wb, file = "C://Documents//publication.xlsx")
```

You could also open a temporary copy of the workbook with `my_wb$open()`, which is convenient during the development process.

#### Output

The content of your output spreadsheet will end up looking something like this:

<div class="figure">
<img src="../man/figures/vignette-workbook.gif" alt="Gif of a simple Excel workbook created using the aftables package, revealing tabs for cover, contents, notes and tables." width="100%"/>
</div>

You'll notice that various best-practice formatting (e.g. Arial size 12 font for body text) and mark-up (e.g. tables, donated by a marker in the lower-right corner of the lower-right cell of each one) have been applied throughout.

Note also that two 'pre-table' meta-elements were created automatically in the sheets that contain statistical tables, which you didn't need to supply to `create_aftable()`: (1) the number of tables and (2) the presence of notes. These are required for accessibility good practice.

#### Final tweaks

It's your responsibility to check and amend the output from {aftables} to ensure it meets users' accessibility needs.

You can apply some final tweaks to the output xlsx file if the defaults don't quite meet your requirements (e.g. some column widths), though it's advisable to keep changes to a minimum for reproducibility purposes and because you may undo some of the compliant structuring and formatting that {aftables} provides.

You will need to manually address some other [accessibility requirements](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/) that are not yet covered by the package such as converting the output to an ODS file rather than the proprietary xlsx format.

We're trying to address a number of these limitations. Please [see the issues on GitHub](https://github.com/best-practice-and-impact/aftables/issues) for the current status.

## Contribute

To contribute, please add [an issue](https://github.com/best-practice-and-impact/aftables/issues) or [a pull request](https://github.com/best-practice-and-impact/aftables/pulls) after reading [the code of conduct](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CODE_OF_CONDUCT.md) and [contributing](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CONTRIBUTING.md) guidance. 

---
title: "Spreadsheet terminology"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Spreadsheet terminology}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)
```

# Purpose

This vignette provides a quick lookup of spreadsheet nomenclature, as used in the {aftables} package.

# Terminology

Certain language and conventions are used in the package to talk about spreadsheets.

A _workbook_ is a spreadsheet-like structure. It contains _tabs_ that are each named with a unique _tab title_. Each tab contains a _sheet_.

Sheets can be one of four _sheet types_ that dictate their content, layout and style:

1. A _cover_ sheet contains the title of the _workbook_ and information about the data it contains, who has produced it, etc
2. A _contents_ sheet contains a table showing the contents of the _workbook_ at a glance, with one row per _sheet_ (not including the cover or itself)
3. A _notes_ sheet contains a table with a lookup of note codes (e.g. '[note 1]') to their explanations
4. One or more _tables_ sheets contain statistical tables (the main purpose for the existence of the spreadsheet) or annexes of supporting information

The _cover_, _contents_ and _notes_ sheets can be considered _meta sheets_ because they provide contextual information about the _workbook_ and its contents.

_Sheets_ themselves are composed of inserted _elements_ that appear in the following row order (if present):

* a _sheet title_ to be displayed at the top of the sheet (user-provided)
* a _table count_ so users know how many tables are in the sheet (auto-generated)
* a _notes statement_ that declares if a table contains _notes_ (if applicable, auto-generated)
* a _blank cells statement_ that indicates the meaning behind any blank cells in a table (if applicable, user-provided)
* a _source statement_ to explain where the data came from (if applicable, user-provided)
* a _table_ that contains information as rows and columns (user-provided) and has a _table name_ (auto-generated) which appears as the 'name' of the marked-up table

# Example

This is an example xlsx output from the {aftables} package:

<div class="figure">
<img src="../man/figures/table-sheet-markup.png" alt="A spreadsheet produced by the aftables package. Four things have been marked with numbers. 1 points to the current tab, named Table_1. 2 points to the sheet title. 3 points to four rows of elements below the title and below the table, which refer to the number of tables in the sheet, the presence of notes, the meaning of blank cells and the data source. 4 points to the table." width="100%"/>
</div>

The labels in the image above highlight:

1. The _tabs_ and their _tab titles_, open on a _sheet_ called 'Table_1' (with each of the _meta sheets_ currently hidden).
2. The _sheet title_.
3. Several _elements_: the _table count_, the _notes statement_ (because the table contains notes), the _blank cells statement_ (because the table contains blank cells) and a _source statement_ for the data.
4. A marked-up _table_, which contains suppressed values (i.e. '[c]') and notes (e.g. '[note 1]').

# Contribute

To contribute, please add [an issue](https://github.com/best-practice-and-impact/aftables/issues) or [a pull request](https://github.com/best-practice-and-impact/aftables/pulls) after reading [the code of conduct](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CODE_OF_CONDUCT.md) and [contributing](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CONTRIBUTING.md) guidance. 

---
title: "Config files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Config files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)
```

## Purpose

This vignette provides an overview of how to use a config file with the {aftables} package to set workbook properties and formatting.

## Workbook properties and formatting which can be set in a config file

Workbook properties:

 - Author
 - Title
 - Keywords
 - Subject
 - Category
 - Comments

Analysis Function guidance advises workbooks should have the author, title, keywords and language document properties completed. aftables provides functionality to set the author, title and keywords properties. See ['Releasing statistics in spreadsheets'](https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/) for more information including how to set the workbook language.

Workbook formatting:

 - Font name
 - Font sizes
 - Column widths

## Structure of a config file

The config file must include an `aftables` entry. {aftables} will only look for workbook configs below the `aftables` entry. The config file may include any other entries. This means {aftables} settings can be added to an existing config file without conflicts.

The `aftables` config file supports multiple entries. There must be at least one config below the `aftables` entry, which could be named `default` or be a user-specified custom config.

Each config must contain at least one of `workbook_properties` and `workbook_format`. The value of each key below `workbook_properties` must be a character string apart from `keywords`, which is a character vector. The value of every key below `workbook_format` must be an integer apart from `base_font_name` which must be a character string. {aftables} checks all config keys to ensure they are character strings or integers as required to process them correctly.

This is the contents of the example config file included with {aftables}:

```{r results='asis', echo = FALSE}
cat("```xml\n")
cat(readLines(system.file("ext-data",
                          "config.yaml",
                          package = "aftables")),
    sep = "\n")
cat("\n```\n")
```

Values of keys below `workbook_properties` will appear in the Excel workbook when it is saved using wb_save from openxlsx2. They can be found in the file properties or the workbook information.

Values of keys below `workbook_format` will be applied to the contents of the workbook. The values of `base_font_name` and `base_font_size` define the default font name and size used by the workbook. All text not formatted as a table header, sheet subheading or sheet heading will use the default settings. Font sizes of sheet headings, sheet subheadings, and table header rows will use the values of `sheet_heading_size`, `sheet_subheading_size` and `table_header_size` respectively, and they will additionally be formatted as bold.

The values of `cellwidth_default`, `cellwidth_wider` and `nchar_break` are used to define column widths. The units of `cellwidth_default` and `cellwidth_wider` are the column width values used by Excel. All columns widths are set by default to use the `cellwidth_default` value. If the number of characters in a column header or the contents of a column exceeds the value of `nchar_break` aftables will set the column width to the value of `cellwidth_wider`. Users can avoid text wrapping in columns or column headers by setting the value of `nchar_break` based on their data or the content of their column headers.

The `create_config_yaml()` function outputs a copy of this config file to a location of the users choice. The default location is the current working directory. We recommend using the example config file as a basis for developing your own {aftables} config file.

## Applying a config file with `generate_workbook()`

The `generate_workbook()` function has two arguments for config files: `config_path` and `config_name`. Both arguments are optional, and {aftables} will take different action based on which arguments are provided.

{aftables} will look for a config file using the `config_path` argument. By default {aftables} will look in the current working directory for a config file named `config.yaml`. `config_path` can be set to NULL and {aftables} will not look for a config file.

{aftables} will only use the `config_name` argument if a config file is found. {aftables} config files must have an `aftables` entry containing one or more configs. These configs can be a `default` config and/or any user-specified custom configs. Users can set up multiple custom configs to produce multiple workbooks from a single config file. Settings common to all workbooks would be set in the `default` config, and settings specific to each workbook would be set in separate custom configs. The `default` config is optional, and custom configs can be used to specify all workbook properties and formatting. When a config file contains both a `default` config and a user-specified custom config matching the `custom_name` argument a combined config is generated from the `default` config and the `config_name` custom config. Values specified in the `config_name` custom config are preferred over the `default` config.

### Example 1: No config arguments in generate_workbook

```{r config example 1, eval = FALSE}
my_wb <- generate_workbook(aftable = demo_aftable)
```

When the user does not provide any config arguments to {aftables} it will look in the working directory for a `config.yaml` file. If {aftables} finds a config file it will look for an `aftables` entry with a `default` config. If {aftables} cannot find a config file or the config file doesn't contain an `aftables` entry with a `default` config, it will generate an {openxlsx2} wbWorkbook object named `my_wb` without applying any config. Otherwise {aftables} will generate an {openxlsx2} wbWorkbook object named `my_wb` with the `default` config applied. This is the default behaviour of the {generate_workbook} function.

### Example 2: Using a custom config

```{r config example 2, eval = FALSE}
my_wb <- generate_workbook(
  aftable = demo_aftable,
  config_name = "workbook1"
)
```

{aftables} will look in the working directory for a file named `config.yaml` and import the `aftables` entry. It will look for the `workbook1` config as well as a `default` config. If there is no `aftables` entry or `workbook1` config, {aftables} will stop with an error. If both the `default` and `workbook1` configs exist {aftables} will combine them, preferring values specified in the `config_name` custom config over those in the `default` config. If there is no `default` config all values will be taken from the `workbook1` config. {aftables} will generate an {openxlsx2} wbWorkbook object named `my_wb` with the `workbook1` or combined config applied.

### Example 3: Using a custom config file

```{r config example 3, eval = FALSE}
dir.create("configs")

create_config_yaml(
  path = "configs",
  open_config = FALSE
)

my_wb <- generate_workbook(
  aftable = demo_aftable,
  config_path = "configs/config.yaml"
)
```

{aftables} will look for the user-specified config file from `config_path` and import the `aftables` entry. It will look for the `default` config. If there is no `aftables` entry or `default` config, {aftables} will stop with an error. Otherwise {aftables} will generate an {openxlsx2} wbWorkbook object named `my_wb` with the `default` config applied.

### Example 4: Using a custom config file with a custom config

```{r config example 4, eval = FALSE}
dir.create("configs")

create_config_yaml(
  path = "configs",
  open_config = FALSE
)

my_wb <- generate_workbook(
  aftable = demo_aftable,
  config_path = "configs/config.yaml",
  config_name = "workbook1"
)
```

{aftables} will look for the user-specified config file from `config_path` and import the `aftables` entry. It will look for the `workbook1` config as well as a `default` config. If there is no `aftables` entry or `workbook1` config, {aftables} will stop with an error. If both the `default` and `workbook1` configs exist {aftables} will combine them, preferring values specified in the `config_name` custom config over those in the `default` config. If there is no `default` config all values will be taken from the `workbook1` config. {aftables} will generate an {openxlsx2} wbWorkbook object named `my_wb` with the `workbook1` or combined config applied.

---
title: "Package structure"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Package structure}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)
```

# Purpose

This vignette is aimed at developers who want to understand the package better and to make it easier for them to contribute.

# Overview

There are only two main user-facing functions in {aftables}: 

* `create_aftable()` to create a data.frame object (with an additional 'aftable' S3 class) filled with all the information needed to create a spreadsheet output, as well as check the validity of the structure and provide errors or warnings
* `generate_workbook()` to convert the output from `create_aftable()` to an [{openxlsx2}](https://janmarvin.github.io/openxlsx2/) wbWorkbook-class object, ready for writing to an xlsx file with `openxlsx2::wb_save()`

This simplicity is a feature, not a bug. It's designed to greatly simplify the process of creating compliant spreadsheets. The package does the hard work of making the outputs compliant so the user spends less time dealing with it.

This vignette provides a quick look at what's happening 'under the hood' in these functions.

Please add [an issue](https://github.com/best-practice-and-impact/aftables/issues) to the package's GitHub repository if you would like any of this explanation to be expanded, or provide a solution in [a pull request](https://github.com/best-practice-and-impact/aftables/pulls).

# Files

First it's worth explaining how the source files are laid out. There are four major groups of scripts in the `R/` directory of the package:

1. Code to make aftable-class objects: `aftable.R` and `utils-aftable.R` contain code for handling the aftable class, most importantly the `create_aftable()` function, but also coercion with `as_aftable()`, checking with `is_aftable()`, a `summary()` method and a `print()` method, which takes advantage of [the {pillar} package](https://github.com/r-lib/pillar) for prettier outputs.
2. Code to make Workbook-class objects: `workbook.R`, `utils-workbook.R` and `utils-workbook-style.R` contain the code for creating and styling a Workbook-class object with the `generate_workbook()` function.
3. Code to produce demo datasets: `data.R` contains the documentation for demo datasets, which are created in the `data-raw/` directory with the files stored in the `data/` directory.
4. Code that creates the RStudio Addin: `addin.R` and `utils-addin.R` contain code for the [RStudio Addin](https://rstudio.github.io/rstudioaddins/) (the .dcf file for which is in the `inst/rstudio/` directory).

You'll also find the `aftables-package.R` file in the `R/` directory, which provides a package-level help page derived from the DESCRIPTION file when `?aftables` is run by the user. It doesn't need to be edited.

# Code

This sections below focus on the `create_aftable()` and `generate_workbook()` functions, which are the primary and most complex functions in the package.

The code that underpins these functions is modularised to aid with bug-catching and testing, but also to make it easier for developers to understand how the code fits together. Internal sub-functions are consistently-named and begin with verbs, which should help you better understand their purpose.

Note that {aftables} uses a convention that internal functions (i.e. those not presented to the user, but accessed via the `:::` qualifier) are prefixed with a period (i.e. `.f()`) to make it clearer that they are internal to the package. The exported user-facing functions do not use a leading period.

## To create aftables

Actually, `create_aftable()` itself only does one thing: it takes user inputs from the arguments and combines them into a dataframe. It then passes this off to the most important function in the package, `as_aftable()`, which is responsible for coercing the dataframe to aftable class and performing checks on its content.

Basically, `as_aftable()` creates an S3-class object with classes 'data.frame' and 'tbl' (i.e. [tibble](https://tibble.tidyverse.org/)) and an additional 'aftable' class.

```{r class-demo}
library(aftables)
my_aftable <- as_aftable(demo_df)
class(my_aftable)
```

The object can be manipulated like a 'normal' dataframe and---thanks to the {pillar} package and the tbl class---it can be printed in compact form without the need for the whole of the {tibble} package to be imported.

```{r print-demo}
my_aftable
```

Compare this to its appearance as a regular data.frame, which is trickier to understand:

```{r df-demo-no-eval}
as.data.frame(my_aftable)
```

Within `as_aftable()` itself are two major functions that help ensure proper construction of an aftable object:

* `.validate_aftable()` will generate errors if basic structural expectations of an aftable aren't met (e.g. if 'cover', 'contents' or 'notes' have been provided more than once to the `sheet_type` argument)
* `.warn_aftable()` checks for things that the user may have forgotten and prints warnings about them (e.g. if 5 notes are declared in the notes sheet but there are fewer in the tables themselves)

Advanced users can create a correctly-formatted data.frame on the fly and convert it to an aftable with `as_aftable()` directly. The `as_aftable()` function mainly exists to make testing easier, i.e. you can pass to it the pre-prepared `demo_df` dataset.

### Methods

There's a few methods for aftables that are also found in `R/aftables.R`.

`is_aftable()` is a classic logical test that checks for the aftable class in the object provided to it.

```{r is-demo}
is_aftable(my_aftable)
```

The `summary()` method prints a very simple overview of a provided aftable.

```{r summary-demo}
summary(my_aftable)
```

The `tbl_sum()` method is provided via the {pillar} package, with the goal of providing a bespoke header to the printed aftable.

```{r tbl-sum-demo}
pillar::tbl_sum(my_aftable)
```

## To create workbooks

The `generate_workbook()` function sets up an [{openxlsx2}](https://janmarvin.github.io/openxlsx2/) wbWorkbook-class object and fills it by iterating over a user-supplied the aftable-class object.

```{r wb-class}
my_wb <- generate_workbook(my_aftable)
class(my_wb)
```

You can see how the Workbook-class object carries information that will determine the structure and style of the final spreadsheet output.

```{r print-wb}
my_wb
```

Several internal sub-functions within `generate_workbook()`---`.add_*()`, `.insert_*()` and `.style_*()`---are responsible for adding these sheets, inserting sheet elements and styling them, respectively.

### Add sheets

A Workbook-class object is first created with `openxlsx2::wb_workbook()` and then sheets are added based on the contents of the user-supplied aftable. 

The following functions add sheets and sheet elements into the workbook:

* `.add_tabs()` adds the required number of tabs into the workbook with `openxlsx2::wb_add_worksheet()` (as per the `tab_title` column of the supplied aftable)
* `.add_cover()` and `.add_contents()` add the information needed for the cover and contents sheets (as per the required 'cover' and 'contents' supplied in the `sheet_type` column of an aftable)
* `.add_notes()` if a notes sheet exists (i.e. a row in the supplied aftable with a `sheet_type` of 'notes')
* `.add_table()` adds sheets for each statistical table (as per rows of supplied aftable with a `sheet_type` of 'table')

As sheets are added, content is inserted and styles are applied with the:

* `.insert_*()` functions, which insert sheet elements (title, source statement, table, etc) to each sheet
* `.style_*()` functions, which apply formatting to each sheet (e.g. bold sheet titles with larger font) and the workbook (e.g. Arial font)

#### Insert sheet elements

There are several `.insert_*()` functions that add information to each sheet depending on the `sheet_type` of the provided aftable, as well as the content, if any, of its `sheet_title`, `blank_cells`, `source` and `table` columns.

The following functions insert 'pre-table' elements in this order:

* `.insert_title()` to place the sheet title in cell A1
* `.insert_table_count()` to add a statement about the number of tables in the sheet
* `.insert_notes_statement()` if a `sheet_type` of 'notes' is provided in the user's aftable
* `.insert_blanks_message()` if content is provided in the `blanks_cells` column of the user's aftable
* `.insert_custom_rows()` if content is provided in the `custom_rows` column of the user's aftable
* `.insert_source()` if content is provided in the `source` column of the user's aftable
    
A table of data is added under the metadata with `.insert_table()`, which is provided in the `table` column of the user's aftable object.

The exact `.insert_*()` functions called depend on the `sheet_type` declared in the aftable:

* meta sheets (cover, contents and notes) need only `.insert_title()` and `.insert_table_count()`
* statistical tables may also require `.insert_blanks_message()`, `.insert_custom_rows()` and `.insert_source()` if the relevant content is provided by the user, as well as `.insert_notes_statement()` if there are notes

Simple logic is used to check for the presence of meta elements with the `.has_*()` functions, while the `get_start_row_*()` functions handle the cell to which each message should be inserted. 

For example, if all the elements are supplied, then the table would begin in row 6 (i.e. after the sheet title, table count, note presence, meaning of blank cells and source), but it's possible that the table would have to be inserted to row 3 if only the sheet title and statement are required. This avoids inaccessible blank rows and redundant statements like 'This table has no source statement'.

#### Apply styles

There are a few `.style_*()` functions that create styles and apply them on the basis of the `sheet_type` provided in the aftable.

* `.style_paragraph()` creates an easily-referenced lookup of paragraph styles.
* `.style_font()` creates an easily-referenced lookup of font styles
* `.style_workbook()` applies defaults for _the whole workbook_ (i.e. to set the font style to Arial size 12)
* `.style_cover()`, `.style_contents()` and `.style_notes()` all apply styles to specific _sheets_
* `.style_sheet_title()` and `.style_table()` apply styles to _particular sheet elements_ (e.g. the title is larger and bolder than the default font)

# Contribute

To contribute, please add [an issue](https://github.com/best-practice-and-impact/aftables/issues) or [a pull request](https://github.com/best-practice-and-impact/aftables/pulls) after reading [the code of conduct](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CODE_OF_CONDUCT.md) and [contributing](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CONTRIBUTING.md) guidance. 

---
title: "Accessibility guidance checklist"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Accessibility guidance checklist}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#"
)
```

## Purpose

This vignette shows you how {aftables} meets accessibility best practice.

## Checklists

This page shows how the features of the {aftables} package meet the Analysis Function's ['Making spreadsheets accessible: checklist of the basics'](https://analysisfunction.civilservice.gov.uk/policy-store/making-spreadsheets-accessible-a-brief-checklist-of-the-basics/) guidance.

The assessments in the table below are based on generating an unedited output from a typical {aftables} workflow, i.e. use `create_aftable()` to make an aftable that's passed to `generate_workbook()` and then written with `openxlsx2::wb_save()`.

There are several self-assessed indicators used in the 'status' column of the tables below:

* 'met' means that {aftables} contains a feature (or the absence of a feature) to help meet the requirement
* 'user's responsibility' means that the user must make a decision about how to meet the requirement
* 'partly' means that {aftables} contains a feature to help meet the requirement, but the user may need to provide additional input
* 'not applicable' means the checklist item does not need to be met due to the way in which {aftables} works

### Tables

| Description | Essential? | Status | Explanation |
| :-- | :- | :- | :---- |
| Mark up tables | Yes | Met | Provided automatically by `generate_workbook()` |
| Give tables meaningful names | No | Met | Unique table names are generated from the `tab_titles` argument in `create_aftable()` |
| Remove merged cells, split cells and nested tables | Yes | Met | {aftables} doesn't create these features |
| Remove blank rows and columns within tables | Yes | Partly/user’s responsibility | A warning is supplied to a user if cells in the table are blank but no explanation has been provided in the `blank_cells` column of the provided aftables object |
| All tables should have one tagged header row | Yes | Met | R's data.frame class only allows for one header row, which must have unique names; the header row is tagged when marked-up as a table |
| Wrap text within cells | Yes | Met | `generate_workbook()` wraps content by default; there's a simple check in {aftables} to widen a column if it contains long strings. |
| Avoid adding filters and freeze panes | No | Met | Filters and freezing aren't supported by {aftables} |
| Only leave cells with no data empty in certain circumstances | Yes | User's responsibility | The user should describe why there are blank cells in the `blank_cells` column of their aftables object |
| Avoid hiding rows or columns | No | Met | Hiding isn't supported by {aftables} |
| Columns should be a sensible width | No | Partly/user's responsibility | {aftables} uses a fixed default column width for data tables (16), which is doubled (32) if the length of the content exceeds a threshold number of characters (50). Users can set these values by using a config file. |

### Footnotes

| Description | Essential? | Status | Explanation |
| :-- | :- | :- | :---- |
| Do not use symbols or superscript to signpost to notes | Yes | Met/user’s responsibility | Notes are detected automatically as digits in square brackets; superscript is not supported by {aftables} |
| Use the word 'note' when referring to footnotes | No | User’s responsibility | Notes are detected in the form '[note 1]'; the user should be consistent about this wording |
| Avoid putting note markers in specific cells | No | User’s responsibility | Columns named 'Notes' and note markers in column headers are detected, the user should not put these elsewhere  |
| Put note text in a notes table on a notes worksheet | No | Met | `create_aftable()` expects the `sheet_type` 'notes' |

### Formatting

| Description | Essential? | Status | Explanation |
| :-- | :- | :- | :---- |
| All written content needs to meet the accessibility guidelines | No | User's responsibility | {aftables} does not check the validity of user-supplied text |
| Links must be accessible | Yes | User's responsibility | {aftables} does not check the validity of user-supplied text |
| Format text to make it accessible | No | Met | `generate_workbook()` auto-formats the text |
| All worksheets should have descriptive titles which are properly tagged and formatted | Yes | Partly/user’s responsibility | {aftables} does not check user-supplied text nor data; the package handles formatting but not yet mark-up for headings |
| Avoid using symbols in general | No | User's responsibility | {aftables} does not check the validity of user-supplied text |
| Do not use headers and footers, floating text boxes or floating toolbars | Yes | Met  | These features aren't supported by {aftables} |
| Do not use visual devices to divide data regions | No | Met | Colours and patterns aren't supported by {aftables} |
| Do not use a background fill | No | Met | Fills aren't supported by {aftables} |
| Do not use colour as the only way to convey a message | Yes | Met | Colour is not supported by {aftables} |
| When using colour for emphasis check the contrast | Yes | Not applicable | Colour is not supported by {aftables}, so a check is not required |
| Avoid images in spreadsheets | No | Met | Images aren't supported by {aftables} |
| Remove macros | No | Met | Macros aren't supported by {aftables} |

### Structure

| Description | Essential? | Status | Explanation |
| :-- | :- | :- | :---- |
| Give worksheets unique names or numbers | Yes | User's responsibility | Supply `tab_titles` to `create_aftable()`, which gives a warning if any names match |
| Remove blank worksheets | Yes | Met | `create_aftable()` will error if expected argument content is missing |
| Use cells in column A wisely | Yes | Met | `generate_workbook()` inserts the sheet title, table count, notes statement, source statement and tables into column A by default  |
| Position tables against the left-hand edges of each sheet | Yes | Met | `generate_workbook()` inserts automatically the tables into column A |
| Avoid putting content below a table | No | Met | `generate_workbook()` adds information to rows above tables only |
| Avoid worksheets with multiple tables | No | Met | The package supports only one table per sheet (for now) |

### Before publishing

| Description | Essential? | Status | Explanation |
| :-- | :- | :- | :---- |
| Run a spelling and grammar check | Yes | User's responsibility | These checks aren't supported by {aftables}; the user should check using their own software |
| Use the accessibility checker | No | User's responsibility | These checks aren't supported by {aftables}; the user should check using their own software |
| Add document information | Yes | Partly/user's responsibility | {aftables} supports setting most document properties by providing arguments or a config file to `generate_workbook()`. Setting document language is not currently supported by {aftables} and should be set by users. |
| Ensure the cursor is in cell A1 of the first worksheet when doing your final save | Yes | Met | Occurs automatically |

# Contribute

To contribute, please add [an issue](https://github.com/best-practice-and-impact/aftables/issues) or [a pull request](https://github.com/best-practice-and-impact/aftables/pulls) after reading [the code of conduct](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CODE_OF_CONDUCT.md) and [contributing](https://github.com/best-practice-and-impact/aftables/blob/main/.github/CONTRIBUTING.md) guidance. 

# tbl output looks as intended

    # aftable: 3 x 7
      tab_title sheet_type sheet_title blank_cells source custom_rows table         
      <chr>     <chr>      <chr>       <chr>       <chr>  <list>      <list>        
    1 A         cover      A           <NA>        <NA>   <chr [1]>   <df [1 x 1]>  
    2 B         contents   B           <NA>        <NA>   <chr [1]>   <df [1 x 2]>  
    3 C         tables     C           <NA>        Source <chr [1]>   <df [32 x 11]>

# string_create_aftable skeleton is okay

    [1] "my_aftable <-\n  aftables::create_aftable(\n    tab_titles = c(\"Cover\", \"Contents\", \"Notes\", \"Table_1\", \"Table_2\"),\n    sheet_types = c(\"cover\", \"contents\", \"notes\", \"tables\", \"tables\"),\n    sheet_titles = c(\n      \"The 'aftables' Demo Workbook\",\n      \"Table of contents\",\n      \"Notes\",\n      \"Table 1: First Example Sheet\",\n      \"Table 2: Second Example Sheet\"\n    ),\n    blank_cells = c(\n      rep(NA_character_, 3),\n      \"Blank cells indicate that there's no note in that row.\",\n      NA_character_\n    ),\n    custom_rows = list(\n      NA_character_,\n      \"A custom row in the Contents sheet.\",\n      NA_character_,\n      c(\n        \"First custom row for Table 1.\",\n        \"A second custom row [with a hyperlink.](https://best-practice-and-impact.github.io/aftables/)\"\n      ),\n      \"A custom row for Table 2\"\n    ),\n    sources = c(\n      rep(NA_character_, 3),\n      \"[The Source Material, 2024](https://best-practice-and-impact.github.io/aftables/)\",\n      \"The Source Material, 2024\"\n    ),\n    tables = list(cover_list, contents_df, notes_df, table_1_df, table_2_df)\n  )"

# string_tables_tibble skeleton is okay

    [1] "cover_list <- list(\n  \"Section 1\" = c(\"First row of Section 1.\", \"Second row of Section 1.\"),\n  \"Section 2\" = \"The only row of Section 2.\",\n  \"Section 3\" = c(\n    \"[Website](https://best-practice-and-impact.github.io/aftables/)\",\n    \"[Email address](mailto:fake.address@aftables.com)\"\n  )\n)\n\ncontents_df <- data.frame(\n  \"Sheet name\" = c(\"Notes\", \"Table 1\", \"Table 2\"),\n  \"Sheet title\" = c(\n    \"Notes used in this workbook\",\n    \"First Example Sheet\",\n    \"Second Example Sheet\"\n  ),\n  check.names = FALSE\n)\n\nnotes_df <- data.frame(\n  \"Note number\" = paste(\"[note \", 1:2, \"]\"),\n  \"Note text\" = c(\"First note.\", \"Second note.\"),\n  check.names = FALSE\n)\n\ntable_1_df <- data.frame(\n  Category = LETTERS[1:10],\n  Numeric = 1:10,\n  \"Numeric suppressed\" = c(1:4, \"[c]\", 6:9, \"[x]\"),\n  \"Numeric thousands\" = abs(round(rnorm(10), 4) * 1e5),\n  \"Numeric decimal\" = abs(round(rnorm(10), 5)),\n  \"Long name that means that the column width needs to be widened\" = 1:10,\n  Notes = c(\"[note 1]\", rep(NA_character_, 4), \"[note 2]\", rep(NA_character_, 4)),\n  check.names = FALSE\n)\n\ntable_2_df <- data.frame(Category = LETTERS[1:10], Numeric = 1:10)"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aftable.R
\name{tbl_sum.aftable}
\alias{tbl_sum.aftable}
\title{Provide A Succinct Summary Of An 'aftable' Object}
\usage{
\method{tbl_sum}{aftable}(x, ...)
}
\arguments{
\item{x}{An aftable-class object to summarise.}

\item{...}{Other arguments to pass.}
}
\value{
Named character vector.
}
\description{
A brief text description of an aftable-class object.
}
\examples{
# Print with description
print(demo_aftable)

# Print description only (package 'tibble' must be installed)
tibble::tbl_sum(demo_aftable)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addin.R
\name{at_template_workflow}
\alias{at_template_workflow}
\title{Insert full 'aftables' template workflow}
\usage{
at_template_workflow()
}
\value{
Empty list. Function is used for side effect.
}
\description{
Insert at the cursor (a) demo templates for cover, contents and notes
tables, and (b) a call to \code{\link{create_aftable}} pre-filled with
demo data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workbook.R
\name{generate_workbook}
\alias{generate_workbook}
\title{Generate A Workbook Object From An 'aftable'}
\usage{
generate_workbook(
  aftable,
  author = NULL,
  title = NULL,
  keywords = NULL,
  config_path = "config.yaml",
  config_name = NULL
)
}
\arguments{
\item{aftable}{An aftable-class object created using
\code{\link{create_aftable}} (or \code{\link{as_aftable}}), which
contains the data and information needed to create a workbook.}

\item{author}{Optional character string to set the workbook author.
Default NULL.}

\item{title}{Optional character string to set the workbook title.
Default NULL.}

\item{keywords}{Optional character vector to set the workbook keywords.
Default NULL.}

\item{config_path}{Optional character string containing path to config file.
Defaults to config.yaml file located in working directory.}

\item{config_name}{Optional character string specifying which configuration
to use from config file. Default NULL.}
}
\value{
An openxlsx2 wbWorkbook-class object.
}
\description{
Populate an 'openxlsx2' wbWorkbook-class object with content from an
aftable-class object. In turn, the output can be passed to
\code{\link[openxlsx2]{wb_save}} from 'openxlsx2'
}
\details{
Analysis Function guidance advises workbooks should have the author, title,
keywords and language document properties completed. \code{aftables} provides
functionality to set the author, title and keywords properties.
See \href{https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/#section-15}{Releasing statistics in spreadsheets}
for more information including how to set the workbook language.

A config file can be used to set workbook properties and formatting. See
\code{\link[aftables]{create_config_yaml}} for details of how to create a
config.yaml file. If author, title or keywords are provided in both the
\code{config.yaml} file and in the \code{generate_workbook} arguments, the values provided in the
function arguments are preferred to those provided in the \code{config.yaml} file.
}
\examples{
# Convert an aftable to an openxlsx2 wbWorkbook-class object
# Setting the minimum workbook properties as function arguments
\dontrun{
example_workbook <- generate_workbook(
  demo_aftable,
  author = "Example author",
  title = "example workbook",
  keywords = c("keyword1", "keyword2", "keyword3")
)}

# Use openxlsx2::wb_get_properties to view properties that have been applied
\dontrun{
openxlsx2::wb_get_properties(example_workbook)}

# Save the workbook with openxlsx2::wb_save
\dontrun{
openxlsx2::wb_save(example_workbook, "example_workbook.xlsx")}

# Using config.yaml file to set workbook properties and edit text and cell
# formatting
\dontrun{
example_workbook2 <- generate_workbook(
  demo_aftable,
  config_path = system.file("ext-data", "config.yaml", package = "aftables"),
  config_name = "workbook1"
  )}

# Use openxlsx2::wb_get_properties to view properties that have been applied
\dontrun{
openxlsx2::wb_get_properties(example_workbook2)}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{demo_df}
\alias{demo_df}
\title{A Demo 'data.frame' Object}
\format{
A data.frame with 6 rows and 7 columns:
\describe{
\item{tab_title}{Character. Text to appear on each sheet's tab.}
\item{sheet_type}{Character. The content type for each sheet: 'cover', 'contents', 'notes', or 'tables'.}
\item{sheet_title}{Character. The title that will appear in cell A1 (top-left) of each sheet.}
\item{blank_cells}{Character. An explanation for any blank cells in the table.}
\item{custom_rows}{List-column of character vectors. Additional arbitrary pre-table information provided by the user.}
\item{source}{Character. The origin of the data, if relevant.}
\item{table}{List-column of data.frames (apart from the cover, which is a list) containing the statistical tables.}
}
}
\usage{
demo_df
}
\description{
A pre-created data.frame ready to be converted to an aftables-class object
with \code{\link{as_aftable}} and then an 'openxlsx2' wbWorkbook-class object
with \code{\link{generate_workbook}}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config_utils.R
\name{create_config_yaml}
\alias{create_config_yaml}
\title{Create an aftables config.yaml}
\usage{
create_config_yaml(path = getwd(), open_config = rlang::is_interactive())
}
\arguments{
\item{path}{optional character string containing directory to copy the
config.yaml file. Defaults to current working directory.}

\item{open_config}{optional logical whether to open the copy of config.yaml
for editing in the current R session.}
}
\description{
Copy the example config.yaml file included with aftables to a directory of
the user's choice, and optionally open the file for editing. The config.yaml
file can be passed to the aftables function
\code{\link[aftables]{generate_workbook}}.
}
\details{
If there is an existing config.yaml file in the destination
directory this function will not overwrite it.

Contents of example config.yaml:

\if{html}{\out{<div class="sourceCode xml">}}\preformatted{aftables:
  default:
    workbook_properties:
      author: "aftables"
      title: "aftables example workbook"
      keywords:
        - "aftables1"
        - "aftables2"
        - "aftables3"
      subject: "aftables example subject"
      category: "aftables example category"
      comments: "aftables example comments"
    workbook_format:
      base_font_name: "Arial"
      base_font_size: 12
      table_header_size: 12
      sheet_heading_size: 16
      sheet_subheading_size: 14
      cellwidth_default: 16
      cellwidth_wider: 32
      nchar_break: 50
  workbook1:
    workbook_properties:
      category: "aftables workbook 1 category"
      keywords:
        - "aftablesworkbook1"
  workbook2:
    workbook_properties:
      title: "aftables workbook 2"
      category: "aftables workbook 2 category"
      keywords:
        - "aftablesworkbook2"

}\if{html}{\out{</div>}}

All configurations must be placed below an \code{aftables} key. \code{aftables} should
be followed by a \code{default} key and/or custom keys (e.g. \code{workbook1}).

The \code{default} key settings will be read by all calls to
\code{\link[aftables]{generate_workbook}} which use this config.yml. This
allows you to share settings when generating multiple workbooks in one
script.

Custom key settings (e.g. \code{workbook1}) will only be used by
\code{\link[aftables]{generate_workbook}} when the key is provided as the
\code{config_name} argument. This allows you to specify settings for a specific
workbook. Custom key settings will be preferred over the \code{default} settings.

Keys below workbook_properties will appear in the Excel workbook when it is
saved using \code{\link[openxlsx2]{wb_save}} from openxlsx2. They can be
found in the file properties or the workbook information.

Keys below \code{workbook_format} will be applied to the contents of the workbook.
The values of \code{base_font_name} and \code{base_font_size} define the default font
name and size used by the workbook. All text not formatted as a table header,
sheet subheading or sheet heading will use the default settings. Font sizes
of sheet headings, sheet subheadings, and table header rows will use the
values of \code{sheet_heading_size}, \code{sheet_subheading_size} and
\code{table_header_size} respectively, and they will additionally be formatted as
bold.

The values of \code{cellwidth_default}, \code{cellwidth_wider} and \code{nchar_break} are
used to define column widths. The units of \code{cellwidth_default} and
\code{cellwidth_wider} are the column width values used by Excel. All columns
widths are set by default to use the \code{cellwidth_default} value. If the number
of characters in a column header or the contents of a column exceeds the
value of \code{nchar_break} aftables will set the column width to the value of
\code{cellwidth_wider}. Users can avoid text wrapping in columns or column headers
by setting the value of \code{nchar_break} based on their data or the content of
their column headers.

Not all workbook configuration options need to be set. Required settings are
documented in \code{\link[aftables]{generate_workbook}}.
}
\examples{
# Use default arguments to copy `aftables` `config.yaml` file to the current
# working directory without opening the copied file for editing:

\dontrun{
create_config_yaml(open_config = FALSE)}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addin.R
\name{at_template_aftable}
\alias{at_template_aftable}
\title{Insert 'create_aftable' template}
\usage{
at_template_aftable()
}
\value{
Empty list. Function is used for side effect.
}
\description{
Insert at the cursor a template for \code{\link{create_aftable}} from the
'aftables' package, pre-filled with demo data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{demo_workbook}
\alias{demo_workbook}
\title{A Demo 'Workbook' Object}
\format{
An 'openxlsx2' wbWorkbook-class object with 5 sheets.
}
\usage{
demo_workbook
}
\description{
A pre-created 'openxlsx2' wbWorkbook-class object generated from an
aftables-class object with \code{\link{generate_workbook}}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aftable.R
\name{summary.aftable}
\alias{summary.aftable}
\title{Summarise An 'aftable' Object}
\usage{
\method{summary}{aftable}(object, ...)
}
\arguments{
\item{object}{An aftable-class object for which to get a summary.}

\item{...}{Other arguments to pass.}
}
\value{
\code{object} unaltered.
}
\description{
A concise result summary of an aftable-class object to see information
about the sheet content. Shows a numbered list of sheets with each tab title,
sheet type and table dimensions.
}
\examples{
# Print a concise summary of the aftable-class object
summary(demo_aftable)

# Alternatively, look at the structure
str(demo_aftable, max.level = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aftable.R
\name{as_aftable}
\alias{as_aftable}
\alias{is_aftable}
\title{Coerce To An 'aftable' Object}
\usage{
as_aftable(x)

is_aftable(x)
}
\arguments{
\item{x}{A data.frame object to coerce.}
}
\value{
\code{as_aftable} returns an object of class aftable if possible.
\code{is_aftable} returns \code{TRUE} if the object has class
aftable, otherwise \code{FALSE}.
}
\description{
Functions to check if an object is an aftable, or coerce it if possible.
}
\examples{
as_aftable(demo_df)
is_aftable(demo_aftable)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aftable.R
\name{create_aftable}
\alias{create_aftable}
\title{Create An 'aftable' Object}
\usage{
create_aftable(
  tab_titles,
  sheet_types = c("cover", "contents", "notes", "tables"),
  sheet_titles,
  blank_cells = NA_character_,
  sources = NA_character_,
  custom_rows = list(NA_character_),
  tables
)
}
\arguments{
\item{tab_titles}{Required character vector, one value per sheet. Each title
will appear literally on each tab of the final spreadsheet output. Keep
brief. Letters and numbers only; do not start with a number; use
underscores for spaces. For example: 'Cover', 'Contents', 'Notes',
'Table_1'. Will be corrected automatically unless there's an error.}

\item{sheet_types}{Required character vector, one value per sheet. Sheets
that don't contain publication tables ('meta' sheets) should be of type
'contents', 'cover' or 'notes'. Sheets that contain statistical tables of
data are type 'tables'.}

\item{sheet_titles}{Required character vector, one value per sheet. The main
title for each sheet, which will appear in cell A1 (top-left corner).}

\item{blank_cells}{Optional character vector, one value per sheet. A short
sentence to explain the reason for any blank cells in the sheet. Supply
as \code{NA_character_} if empty. Most likely to be used with sheet type
'tables'.}

\item{sources}{Optional character vector, one value per sheet. The origin of
the data for a given sheet. Supply as \code{NA_character_} if empty. To
be used with sheet type 'tables'.}

\item{custom_rows}{Optional list of character vectors. One list element per
sheet, one character vector element per row of pre-table metadata. Supply
a list element as \code{NA_character_} if empty. To be used with sheet
type 'tables', but can also be used for sheet types 'contents' and
'notes'.}

\item{tables}{Required list of data.frames (though the cover sheet may be
supplied as a list), one per sheet. See details.}
}
\value{
An object with classes 'aftable', 'tbl' and 'data.frame'.
}
\description{
Create a new aftable-class object, which is a special data.frame that
contains all the information needed in your output spreadsheet. In turn, the
object created by this function can be used to populate an 'openxlsx2'
wbWorkbook-class object with the function \code{\link{generate_workbook}}.
}
\details{
\subsection{How to supply data to the 'tables' argument}{

Formats for the elements collected as a list and passed to the 'tables'
argument, depending on the sheet type.

\itemize{
\item Sheet type 'cover': either (a) a list where each element name is a
section header and each element's content is a character vector
whose elements will make up separate rows of that section
(recommended), or (b) a data.frame with one row per subsection, with
one column for section titles and one column for corresponding for
that section's body text. For example, you may have a section with
the title 'Contact details' that contains an email address and
telephone number. You can use line breaks (i.e. '\\n') to separate
text into paragraphs.
\item Sheet type 'contents': one row per sheet, two columns suggested at
least (named 'Tab title' and 'Worksheet title').
\item Sheet type 'notes': one row per note, two columns suggested (named
'Note number', 'Note text'), where notes are in the form
'[note 1]'.
\item Sheet type 'tables': a tidy, rectangular data.frame containing the
data to be published. It's the user's responsibility to add notes in
the form '[note 1]' to column headers, or in a special 'Notes' row.
}
}

\subsection{How to supply hyperlinks}{

You can provide text in Markdown link syntax (e.g.
'[GOV.UK](https://www.gov.uk)', adding 'mailto:' before an
email address) and the containing cell will be rendered as a hyperlink in the
output spreadsheet. Note that whole cells will become hyperlinks; there is no
support for selected words in a sentence to be rendered as a hyperlink.

Hyperlinks can be supplied in the character strings to three arguments:

\itemize{
\item To the 'tables' argument for sheet type 'cover' only. It's
recommended to supply the cover information as a list rather than a
data.frame, which will allow you to make specific rows within a
section (e.g. 'contact us') into hyperlinks.
\item To the 'custom_rows' argument for sheets of type 'contents, 'notes'
and 'tables'.
\item To the 'source' argument for sheets of type 'table' only.
}
}
}
\examples{
# Prepare some demo tables of information

set.seed(1066)

cover_list <- list(
  "Section 1" = c("First row of Section 1.", "Second row of Section 1."),
  "Section 2" = "The only row of Section 2.",
  "Section 3" = c(
    "[Website](https://best-practice-and-impact.github.io/aftables/)",
    "[Email address](mailto:fake.address@aftables.com)"
  )
)

contents_df <- data.frame(
  "Sheet name" = c("Notes", "Table_1", "Table_2"),
  "Sheet title" = c(
    "Notes used in this workbook",
    "First Example Sheet",
    "Second Example Sheet"
  ),
  check.names = FALSE
)

notes_df <- data.frame(
  "Note number" = paste0("[note ", 1:2, "]"),
  "Note text" = c("First note.", "Second note."),
  check.names = FALSE
)


table_1_df <- data.frame(
  Category = LETTERS[1:10],
  "Numeric [note 1]" = 1:10,
  "Numeric suppressed" = c(1:4, "[c]", 6:9, "[x]"),
  "Numeric thousands" = abs(round(rnorm(10), 4) * 1e5),
  "Numeric decimal" = abs(round(rnorm(10), 5)),
  "Long name that means that the column width needs to be widened" = 1:10,
  Notes = c(
    "[note 1]", rep(NA_character_, 4), "[note 2]",
    rep(NA_character_, 4)
  ),
  check.names = FALSE
)

table_2_df <- data.frame(Category = LETTERS[1:10], Numeric = 1:10)

# Create 'aftables' object

x <- aftables::create_aftable(
  tab_titles = c("Cover", "Contents", "Notes", "Table_1", "Table_2"),
  sheet_types = c("cover", "contents", "notes", "tables", "tables"),
  sheet_titles = c(
    "The 'aftables' Demo Workbook",
    "Table of contents",
    "Notes",
    "Table 1: First Example Sheet",
    "Table 2: Second Example Sheet"
  ),
  blank_cells = c(
    rep(NA_character_, 3),
    "Blank cells indicate that there's no note in that row.",
    NA_character_
  ),
  custom_rows = list(
    NA_character_,
    NA_character_,
    "A custom row.",
    c(
      paste0(
        "First custom row [with a hyperlink.]",
        "(https://best-practice-and-impact.github.io/aftables/)"
      ),
      "Second custom row."
    ),
    "A custom row."
  ),
  sources = c(
    rep(NA_character_, 3),
    paste0(
      "[The Source Material, 2024.]",
      "(https://best-practice-and-impact.github.io/aftables/)"
    ),
    "The Source Material, 2024."
  ),
  tables = list(cover_list, contents_df, notes_df, table_1_df, table_2_df)
)

# Test that 'aftable' is one of the object's classes
is_aftable(x)

# Look at the structure of the object
str(x, max.level = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aftables-package.R
\docType{package}
\name{aftables-package}
\alias{aftables}
\alias{aftables-package}
\title{aftables: Create Spreadsheet Publications Following Best Practice}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Generate spreadsheet publications that follow best practice guidance from the UK government's Analysis Function, available at \url{https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/}, with a focus on accessibility. See also the 'Python' package 'gptables'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://best-practice-and-impact.github.io/aftables/}
  \item \url{https://github.com/best-practice-and-impact/aftables}
  \item Report bugs at \url{https://github.com/best-practice-and-impact/aftables/issues}
}

}
\author{
\strong{Maintainer}: Olivia Box Power \email{Olivia.BoxPower@dhsc.gov.uk} [contributor]

Authors:
\itemize{
  \item Matt Dray \email{mwdray@gmail.com}
}

Other contributors:
\itemize{
  \item Tim Taylor [contributor]
  \item Matt Kerlogue [contributor]
  \item Crown Copyright [copyright holder]
  \item Zachary Gleisner \email{Zachary.Gleisner@dhsc.gov.uk} [contributor]
}

}
\keyword{internal}
